""" Provide functions for preprocessing the WAV files. """

import logging
from pathlib import Path
import subprocess
from typing import List

from pydub import AudioSegment # type: ignore

from .. import config
from ..utterance import Utterance

logger = logging.getLogger(__name__) # type: ignore

def millisecs_to_secs(millisecs: int) -> float:
    return millisecs / 1000

def trim_wav_ms(in_path: Path, out_path: Path,
                start_time: int, end_time: int) -> None:
    """ Extracts part of a WAV File.

    First attempts to call sox. If sox is unavailable, it backs off to
    pydub+ffmpeg.

    Args:
        in_path: A path to the source file to extract a portion of
        out_path: A path describing the to-be-created WAV file.
        start_time: The point in the source WAV file at which to begin
            extraction.
        end_time: The point in the source WAV file at which to end extraction.

    """

    try:
        trim_wav_sox(in_path, out_path, start_time, end_time)
    except FileNotFoundError:
        # Then sox isn't installed, so use pydub/ffmpeg
        trim_wav_pydub(in_path, out_path, start_time, end_time)
    except subprocess.CalledProcessError:
        # Then there is an issue calling sox. Perhaps the input file is an mp4
        # or some other filetype not supported out-of-the-box by sox. So we try
        # using pydub/ffmpeg.
        trim_wav_pydub(in_path, out_path, start_time, end_time)

def trim_wav_pydub(in_path: Path, out_path: Path,
                start_time: int, end_time: int) -> None:
    """ Crops the wav file. """

    logger.info(
        "Using pydub/ffmpeg to create {} from {}".format(out_path, in_path) +
        " using a start_time of {} and an end_time of {}".format(start_time,
                                                                 end_time))

    if out_path.is_file():
        return

    # TODO add logging here
    #print("in_fn: {}".format(in_fn))
    #print("out_fn: {}".format(out_fn))
    in_ext = in_path.suffix[1:]
    out_ext = out_path.suffix[1:]
    audio = AudioSegment.from_file(str(in_path), in_ext)
    trimmed = audio[start_time:end_time]
    # pydub evidently doesn't actually use the parameters when outputting wavs,
    # since it doesn't use FFMPEG to deal with outputtting WAVs. This is a bit
    # of a leaky abstraction. No warning is given, so normalization to 16Khz
    # mono wavs has to happen later. Leaving the parameters here in case it
    # changes
    trimmed.export(str(out_path), format=out_ext,
                   parameters=["-ac", "1", "-ar", "16000"])

def trim_wav_sox(in_path: Path, out_path: Path,
                 start_time: int, end_time: int) -> None:
    """ Crops the wav file at in_fn so that the audio between start_time and
    end_time is output to out_fn. Measured in milliseconds.
    """

    if out_path.is_file():
        logger.info("Output path %s already exists, not trimming file", out_path)
        return

    start_time_secs = millisecs_to_secs(start_time)
    end_time_secs = millisecs_to_secs(end_time)
    args = [config.SOX_PATH, str(in_path), str(out_path),
            "trim", str(start_time_secs), "=" + str(end_time_secs)]
    logger.info("Cropping file %s, from start time %d (seconds) to end time %d (seconds), outputting to %s",
                in_path, start_time_secs, end_time_secs, out_path)
    subprocess.run(args, check=True)

def extract_wavs(utterances: List[Utterance], tgt_dir: Path,
                 lazy: bool) -> None:
    """ Extracts WAVs from the media files associated with a list of Utterance
    objects and stores it in a target directory.

    Args:
        utterances: A list of Utterance objects, which include information
            about the source media file, and the offset of the utterance in the
            media_file.
        tgt_dir: The directory in which to write the output WAVs.
        lazy: If True, then existing WAVs will not be overwritten if they have
            the same name
    """
    tgt_dir.mkdir(parents=True, exist_ok=True)
    for utter in utterances:
        wav_fn = "{}.{}".format(utter.prefix, "wav")
        out_wav_path = tgt_dir / wav_fn
        if lazy and out_wav_path.is_file():
            logger.info("File {} already exists and lazy == {}; not " \
                         "writing.".format(out_wav_path, lazy))
            continue
        logger.info("File {} does not exist and lazy == {}; creating " \
                     "it.".format(out_wav_path, lazy))
        trim_wav_ms(utter.org_media_path, out_wav_path,
                    utter.start_time, utter.end_time)



""" Miscellaneous functions relating to reporting experimental results."""

from pathlib import Path
from typing import Set, Dict, Tuple, Sequence, List, Union

from collections import Counter
from . import utils

from .distance import min_edit_distance_align

def filter_labels(sent: Sequence[str], labels: Set[str] = None) -> List[str]:
    """ Returns only the tokens present in the sentence that are in labels."""

    if labels:
        return [tok for tok in sent if tok in labels]
    return list(sent)

def filtered_error_rate(hyps_path: Union[str, Path], refs_path: Union[str, Path], labels: Set[str]) -> float:
    """ Returns the error rate of hypotheses in hyps_path against references in refs_path after filtering only for labels in labels.
    """
    if isinstance(hyps_path, Path):
        hyps_path = str(hyps_path)
    if isinstance(refs_path, Path):
        refs_path = str(refs_path)

    with open(hyps_path) as hyps_f:
        lines = hyps_f.readlines()
        hyps = [filter_labels(line.split(), labels) for line in lines]
    with open(refs_path) as refs_f:
        lines = refs_f.readlines()
        refs = [filter_labels(line.split(), labels) for line in lines]

    # For the case where there are no tokens left after filtering.
    only_empty = True
    for entry in hyps:
        if entry is not []:
            only_empty = False
            break # found something so can move on immediately
    if only_empty:
        return -1

    return utils.batch_per(hyps, refs)

def latex_header() -> str:
    """ Produces a LaTeX header for the fmt_latex_*() functions to use. """

    return (r"""\documentclass[10pt]{article}
\usepackage[a4paper,margin=0.5in,landscape]{geometry}
\usepackage[utf8]{inputenc}
\usepackage{xcolor}"
\usepackage{polyglossia}
\usepackage{booktabs}
\usepackage{longtable}
\setmainfont[Mapping=tex-text,Ligatures=Common,Scale=MatchLowercase]{Doulos SIL}"
\DeclareRobustCommand{hl}[1]{{\textcolor{red}{#1}}}""")

def fmt_latex_output(hyps: Sequence[Sequence[str]],
                     refs: Sequence[Sequence[str]],
                     prefixes: Sequence[str],
                     out_fn: Path,
                    ) -> None:
    """ Output the hypotheses and references to a LaTeX source file for
    pretty printing.
    """

    alignments_ = [min_edit_distance_align(ref, hyp)
                  for hyp, ref in zip(hyps, refs)]

    with out_fn.open("w") as out_f:
        print(latex_header(), file=out_f)

        print("\\begin{document}\n"
              "\\begin{longtable}{ll}", file=out_f)
        print(r"\toprule", file=out_f)
        for sent in zip(prefixes, alignments_):
            prefix = sent[0]
            alignments = sent[1:]
            print("Utterance ID: &", prefix.strip().replace(r"_", r"\_"), r"\\", file=out_f)
            for i, alignment in enumerate(alignments):
                ref_list = []
                hyp_list = []
                for arrow in alignment:
                    if arrow[0] == arrow[1]:
                        # Then don't highlight it; it's correct.
                        ref_list.append(arrow[0])
                        hyp_list.append(arrow[1])
                    else:
                        # Then highlight the errors.
                        ref_list.append("\\hl{%s}" % arrow[0])
                        hyp_list.append("\\hl{%s}" % arrow[1])
                print("Ref: &", "".join(ref_list), r"\\", file=out_f)
                print("Hyp: &", "".join(hyp_list), r"\\", file=out_f)
            print(r"\midrule", file=out_f)

        print(r"\end{longtable}", file=out_f)
        print(r"\end{document}", file=out_f)

def fmt_error_types(hyps: Sequence[Sequence[str]],
                    refs: Sequence[Sequence[str]]
                   ) -> str:
    """ Format some information about different error types: insertions, deletions and substitutions."""

    alignments = [min_edit_distance_align(ref, hyp)
                  for hyp, ref in zip(hyps, refs)]

    arrow_counter = Counter() # type: Dict[Tuple[str, str], int]
    for alignment in alignments:
        arrow_counter.update(alignment)
    sub_count = sum([count for arrow, count in arrow_counter.items()
                if arrow[0] != arrow[1] and arrow[0] != "" and arrow[1] != ""])
    dels = [(arrow[0], count) for arrow, count in arrow_counter.items()
            if arrow[0] != arrow[1] and arrow[0] != "" and arrow[1] == ""]
    del_count = sum([count for arrow, count in dels])
    ins_count = sum([count for arrow, count in arrow_counter.items()
                if arrow[0] != arrow[1] and arrow[0] == "" and arrow[1] != ""])
    total = sum([count for _, count in arrow_counter.items()])

    fmt_pieces = []
    fmt = "{:15}{:<4}\n"
    fmt_pieces.append(fmt.format("Substitutions", sub_count))
    fmt_pieces.append(fmt.format("Deletions", del_count))
    fmt_pieces.append(fmt.format("Insertions", ins_count))
    fmt_pieces.append("\n")
    fmt_pieces.append("Deletions:\n")
    fmt_pieces.extend(["{:4}{:<4}\n".format(label, count)
                       for label, count in
                       sorted(dels, reverse=True, key=lambda x: x[1])])

    return "".join(fmt_pieces)


def fmt_confusion_matrix(hyps: Sequence[Sequence[str]],
                         refs: Sequence[Sequence[str]],
                         label_set: Set[str] = None,
                         max_width: int = 25) -> str:
    """ Formats a confusion matrix over substitutions, ignoring insertions
    and deletions. """

    if not label_set:
        # Then determine the label set by reading
        raise NotImplementedError()

    alignments = [min_edit_distance_align(ref, hyp)
                  for hyp, ref in zip(hyps, refs)]

    arrow_counter = Counter() # type: Dict[Tuple[str, str], int]
    for alignment in alignments:
        arrow_counter.update(alignment)

    ref_total = Counter() # type: Dict[str, int]
    for alignment in alignments:
        ref_total.update([arrow[0] for arrow in alignment])

    labels = [label for label, count
              in sorted(ref_total.items(), key=lambda x: x[1], reverse=True)
              if label != ""][:max_width]

    format_pieces = []
    fmt = "{:3} "*(len(labels)+1)
    format_pieces.append(fmt.format(" ", *labels))
    fmt = "{:3} " + ("{:<3} " * (len(labels)))
    for ref in labels:
        # TODO
        ref_results = [arrow_counter[(ref, hyp)] for hyp in labels]
        format_pieces.append(fmt.format(ref, *ref_results))

    return "\n".join(format_pieces)

def fmt_latex_untranscribed(hyps: Sequence[Sequence[str]],
                            prefixes: Sequence[str],
                            out_fn: Path) -> None:
    """ Formats automatic hypotheses that have not previously been
    transcribed in LaTeX. """

    hyps_prefixes = list(zip(hyps, prefixes))
    def utter_id_key(hyp_prefix):
        hyp, prefix = hyp_prefix
        prefix_split = prefix.split(".")
        return (prefix_split[0], int(prefix_split[1]))
    hyps_prefixes.sort(key=utter_id_key)

    with out_fn.open("w") as out_f:
        print(latex_header(), file=out_f)
        print("\\begin{document}\n"
              "\\begin{longtable}{ll}", file=out_f)
        print(r"\toprule", file=out_f)
        for hyp, prefix in hyps_prefixes:
            print("Utterance ID: &", prefix.strip().replace(r"_", r"\_"), "\\\\", file=out_f)
            print("Hypothesis: &", hyp, r"\\", file=out_f)
            print("\\midrule", file=out_f)
        print(r"\end{longtable}", file=out_f)
        print(r"\end{document}", file=out_f)



""" Configuration for some variables that may be system dependent.

Supply a settings.ini file to override values. The format of the file will
have the following sections

[PATHS]
CORPORA_BASE_PATH = /my/path/to/corpora
TARGET = /path/to/preprocessed/data
EXPERIMENTS = /path/to/experiment/output
"""
import configparser
import os
from pkg_resources import Requirement, resource_filename

config_file = configparser.ConfigParser()
config_file.read('settings.ini')

CORPORA_BASE_PATH = config_file.get("PATHS", "CORPORA_BASE_PATH",
                                    fallback="./data/org/")

# The directory of the original source corpora, un-preprocessed.
NA_PATH = config_file.get("PATHS", "NA_PATH", fallback=os.path.join(CORPORA_BASE_PATH, "na"))

# For Kunwinjku data:
BKW_PATH = config_file.get("PATHS", "BKW_PATH",
                           fallback=os.path.join(CORPORA_BASE_PATH, "BKW-speaker-ids_2/"))
EN_WORDS_PATH = config_file.get("PATHS", "EN_WORDS_PATH",
                                fallback=os.path.join(CORPORA_BASE_PATH, "english-words/words.txt"))

# The directory where the preprocessed data will be held.
TGT_DIR = config_file.get("PATHS", "TARGET", fallback="./data")
# The path for experiments
EXP_DIR = config_file.get("PATHS", "EXPERIMENTS", fallback="./exp")

TEST_ROOT = config_file.get("PATHS", "TEST_ROOT", fallback="./testing/")
TEST_EXP_PATH = os.path.join(TEST_ROOT, "exp/")
TEST_DATA_PATH = os.path.join(TEST_ROOT, "data/")

# The path to the sox tool; currently used for splitting WAVs, but we can
# replace with pydub. Actually, the pydub approach is slow so now
# wav.trim_wav_ms tries to use sox and then fallsback to pydub/ffmpeg
SOX_PATH = config_file.get("PATHS", "SOX_PATH", fallback="sox")
# FFMPEG is used for normalizing WAVs
FFMPEG_PATH = config_file.get("PATHS", "FFMPEG_PATH", fallback="ffmpeg")
# Kaldi is used for pitch extraction
KALDI_ROOT = config_file.get("PATHS", "KALDI_ROOT_PATH", fallback="/home/oadams/tools/kaldi")

# Fetch the path of the logging.ini file installed by setuptools.
logging_ini_path = resource_filename(Requirement.parse("persephone"), "persephone/logging.ini")

LOGGING_INI_PATH = config_file.get("PATHS", "log_ini_path", fallback=logging_ini_path)

ENCODING="utf8"



""" Provides a Corpus class that can read ELAN .eaf XML files. """

import logging
from pathlib import Path
from typing import List, Dict, Tuple

import pympi.Elan

from ..utterance import Utterance

logger = logging.getLogger(__name__) # type: ignore

class Eaf(pympi.Elan.Eaf):
    """ This subclass exists because eaf MEDIA_DESCRIPTOR elements typically
    have RELATIVE_MEDIA_URL that contains the path of the media file relative
    to the eaf file. However, pympi.Elan.Eaf doesn't have a path attribute
    pointing to the eaf file itself. This class here adds this attribute so
    that the path to the media file can be reconstructed."""

    def __init__(self, eaf_path: Path) -> None:
        super().__init__(str(eaf_path))
        self.eaf_path = eaf_path
        self.initialize_media_descriptor()

    @property
    def media_path(self) -> Path:
        """ The URL of the associated media file."""
        return self.get_media_path(self.media_descriptor)

    @property
    def time_origin(self) -> int:
        """ The time offset of the annotations."""
        try:
            return int(self.media_descriptor["TIME_ORIGIN"])
        except KeyError:
            # Since it wasn't specified, we don't need to adjust 
            # utterance offsets.
            return 0

    def get_media_path(self, media_descriptor: Dict) -> Path:
        return self.eaf_path.parent / media_descriptor["RELATIVE_MEDIA_URL"]

    def initialize_media_descriptor(self) -> None:
        """
        Returns the media descriptor for the first media descriptor where
        the file can be found.
        """

        for md in self.media_descriptors:
            media_path = self.get_media_path(md)
            if media_path.is_file():
                self.media_descriptor = md
                return

        raise FileNotFoundError(
            """Cannot find media file corresponding to {}.
            Tried looking for the following files: {}.
            """.format(self.eaf_path, [self.get_media_path(md)
                                       for md in self.media_descriptors]))


def sort_annotations(annotations: List[Tuple[int, int, str]]
                     ) -> List[Tuple[int, int, str]]:
    """ Sorts the annotations by their start_time. """
    return sorted(annotations, key=lambda x: x[0])


def utterances_from_tier(eafob: Eaf, tier_name: str) -> List[Utterance]:
    """ Returns utterances found in the given Eaf object in the given tier."""

    try:
        speaker = eafob.tiers[tier_name][2]["PARTICIPANT"]
    except KeyError:
        speaker = None # We don't know the name of the speaker.

    tier_utterances = []

    annotations = sort_annotations(
        list(eafob.get_annotation_data_for_tier(tier_name)))

    for i, annotation in enumerate(annotations):
        eaf_stem = eafob.eaf_path.stem
        utter_id = "{}.{}.{}".format(eaf_stem, tier_name, i)
        start_time = eafob.time_origin + annotation[0]
        end_time = eafob.time_origin + annotation[1]
        text = annotation[2]
        utterance = Utterance(eafob.media_path, eafob.eaf_path, utter_id,
                              start_time, end_time, text, speaker)
        tier_utterances.append(utterance)

    return tier_utterances


def utterances_from_eaf(eaf_path: Path, tier_prefixes: Tuple[str, ...]) -> List[Utterance]:
    """
    Extracts utterances in tiers that start with tier_prefixes found in the ELAN .eaf XML file
    at eaf_path.

    For example, if xv@Mark is a tier in the eaf file, and
    tier_prefixes = ["xv"], then utterances from that tier will be gathered.
    """

    if not eaf_path.is_file():
        raise FileNotFoundError("Cannot find {}".format(eaf_path))

    eaf = Eaf(eaf_path)
    utterances = []
    for tier_name in sorted(list(eaf.tiers)): # Sorting for determinism
        for tier_prefix in tier_prefixes:
            if tier_name.startswith(tier_prefix):
                utterances.extend(utterances_from_tier(eaf, tier_name))
                break
    return utterances


def utterances_from_dir(eaf_dir: Path,
                        tier_prefixes: Tuple[str, ...]) -> List[Utterance]:
    """ Returns the utterances found in ELAN files in a directory.

    Recursively explores the directory, gathering ELAN files and extracting
    utterances from them for tiers that start with the specified prefixes.

    Args:
        eaf_dir: A path to the directory to be searched
        tier_prefixes: Stings matching the start of ELAN tier names that are to
            be extracted. For example, if you want to extract from tiers "xv-Jane"
            and "xv-Mark", then tier_prefixes = ["xv"] would do the job.

    Returns:
        A list of Utterance objects.

    """

    logger.info(
        "EAF from directory: {}, searching with tier_prefixes {}".format(
            eaf_dir, tier_prefixes))

    utterances = []
    for eaf_path in eaf_dir.glob("**/*.eaf"):
        eaf_utterances = utterances_from_eaf(eaf_path, tier_prefixes)
        utterances.extend(eaf_utterances)
    return utterances



__version__ = "0.4.1"

import sys
import logging

def handle_unhandled_exception(exc_type, exc_value, exc_traceback):
    """Handler for unhandled exceptions that will write to the logs"""
    if issubclass(exc_type, KeyboardInterrupt):
        # call the default excepthook saved at __excepthook__
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return
    logger = logging.getLogger(__name__) # type: ignore
    logger.critical("Unhandled exception", exc_info=(exc_type, exc_value, exc_traceback))

sys.excepthook = handle_unhandled_exception



""" An CorpusReader class that interfaces with preprocessed corpora."""
from __future__ import generator_stop

import logging
import logging.config
from pathlib import Path
import pprint
import random
from typing import List, Sequence, Iterator

import numpy as np

from . import utils
from .config import ENCODING
from .exceptions import PersephoneException

logger = logging.getLogger(__name__) # type: ignore

class CorpusReader:
    """ Interfaces to the preprocessed corpora to read in train, valid, and
    test set features and transcriptions. This interface is common to all
    corpora. It is the responsibility of <corpora-name>.py to preprocess the
    data into a valid structure of
    <corpus-name>/[mam-train|mam-valid<seed>|mam-test].  """

    rand = True

    def __init__(self, corpus, num_train=None, batch_size=None, max_samples=None, rand_seed=0):
        """ Construct a new `CorpusReader` instance.

            corpus: The Corpus object that interfaces with a given corpus.
            num_train: The number of training instances from the corpus used.
            batch_size: The size of the batches to yield. If None, then it is
                        num_train / 32.0.
            max_samples: The maximum length of utterances measured in samples.
                         Longer utterances are filtered out.
            rand_seed: The seed for the random number generator. If None, then
                       no randomization is used.
        """

        self.corpus = corpus

        if max_samples:
            logger.critical("max_samples not yet implemented in CorpusReader")
            raise NotImplementedError("Not yet implemented.")

        if not num_train:
            if not batch_size:
                batch_size = 64
            num_train = len(corpus.get_train_fns()[0])
            num_batches = int(num_train / batch_size)
            num_train = int(num_batches * batch_size)
        self.num_train = num_train
        logger.info("Number of training utterances: {}".format(num_train))
        logger.info("Batch size: {}".format(batch_size))
        logger.info("Batches per epoch: {}".format(int(num_train/batch_size)))
        print("Number of training utterances: {}".format(num_train))
        print("Batch size: {}".format(batch_size))
        print("Batches per epoch: {}".format(int(num_train/batch_size)))

        if batch_size:
            self.batch_size = batch_size
            if num_train % batch_size != 0:
                logger.error("Number of training examples {} not divisible"
                             " by batch size {}.".format(num_train, batch_size))
                raise PersephoneException("Number of training examples {} not divisible"
                                          " by batch size {}.".format(num_train, batch_size))
        else:
            # Dynamically change batch size based on number of training
            # examples.
            self.batch_size = int(num_train / 32.0)
            if self.batch_size > 64:
                # I was getting OOM errors when training with 4096 sents, as
                # the batch size jumped to 128
                self.batch_size = 64
            # For now we hope that training numbers are powers of two or
            # something... If not, crash before anything else happens.

            if num_train % self.batch_size != 0:
                logger.error("Number of training examples {} not divisible"
                             " by batch size {}.".format(num_train, self.batch_size))
                raise PersephoneException("Number of training examples {} not divisible"
                                          " by batch size {}.".format(num_train, batch_size))

        random.seed(rand_seed)

        # Make a copy of the training prefixes, randomize their order, and take
        # a subset. Doing random selection of a subset of training now ensures
        # the selection of of training sentences is invariant between calls to
        # train_batch_gen()
        self.train_fns = list(zip(*corpus.get_train_fns()))
        if self.rand:
            random.shuffle(self.train_fns)
        self.train_fns = self.train_fns[:self.num_train]

    def load_batch(self, fn_batch):
        """ Loads a batch with the given prefixes. The prefixes is the full path to the
        training example minus the extension.
        """

        # TODO Assumes targets are available, which is how its distinct from
        # utils.load_batch_x(). These functions need to change names to be
        # clearer.

        inverse = list(zip(*fn_batch))
        feat_fn_batch = inverse[0]
        target_fn_batch = inverse[1]

        batch_inputs, batch_inputs_lens = utils.load_batch_x(feat_fn_batch,
                                                             flatten=False)
        batch_targets_list = []
        for targets_path in target_fn_batch:
            with open(targets_path, encoding=ENCODING) as targets_f:
                target_indices = self.corpus.labels_to_indices(targets_f.readline().split())
                batch_targets_list.append(target_indices)
        batch_targets = utils.target_list_to_sparse_tensor(batch_targets_list)

        return batch_inputs, batch_inputs_lens, batch_targets


    def make_batches(self, utterance_fns: Sequence[Path]) -> List[Sequence[Path]]:
        """ Group utterances into batches for decoding.  """

        return utils.make_batches(utterance_fns, self.batch_size)

    def train_batch_gen(self) -> Iterator:
        """ Returns a generator that outputs batches in the training data."""

        if len(self.train_fns) == 0:
            raise PersephoneException("""No training data available; cannot
                                       generate training batches.""")

        # Create batches of batch_size and shuffle them.
        fn_batches = self.make_batches(self.train_fns)


        if self.rand:
            random.shuffle(fn_batches)

        for fn_batch in fn_batches:
            logger.debug("Batch of training filenames: %s",
                          pprint.pformat(fn_batch))
            yield self.load_batch(fn_batch)
        else:
            # Python 3.7 compatible way to mark generator as exhausted
            return

    def valid_batch(self):
        """ Returns a single batch with all the validation cases."""

        valid_fns = list(zip(*self.corpus.get_valid_fns()))
        return self.load_batch(valid_fns)

    def test_batch(self):
        """ Returns a single batch with all the test cases."""

        test_fns = list(zip(*self.corpus.get_test_fns()))
        return self.load_batch(test_fns)

    def untranscribed_batch_gen(self):
        """ A batch generator for all the untranscribed data. """

        feat_fns = self.corpus.get_untranscribed_fns()
        fn_batches = self.make_batches(feat_fns)

        for fn_batch in fn_batches:
            batch_inputs, batch_inputs_lens = utils.load_batch_x(fn_batch,
                                                             flatten=False)
            yield batch_inputs, batch_inputs_lens, fn_batch

    def human_readable_hyp_ref(self, dense_decoded, dense_y):
        """ Returns a human readable version of the hypothesis for manual
        inspection, along with the reference.
        """

        hyps = []
        refs = []
        for i in range(len(dense_decoded)):
            ref = [phn_i for phn_i in dense_y[i] if phn_i != 0]
            hyp = [phn_i for phn_i in dense_decoded[i] if phn_i != 0]
            ref = self.corpus.indices_to_labels(ref)
            hyp = self.corpus.indices_to_labels(hyp)
            refs.append(ref)
            hyps.append(hyp)

        return hyps, refs

    def human_readable(self, dense_repr: Sequence[Sequence[int]]) -> List[List[str]]:
        """ Returns a human readable version of a dense representation of
        either or reference to facilitate simple manual inspection.
        """

        transcripts = []
        for dense_r in dense_repr:
            non_empty_phonemes = [phn_i for phn_i in dense_r if phn_i != 0]
            transcript = self.corpus.indices_to_labels(non_empty_phonemes)
            transcripts.append(transcript)

        return transcripts

    def __repr__(self):
        return ("%s(" % self.__class__.__name__ +
                "num_train=%s,\n" % repr(self.num_train) +
                "\tbatch_size=%s,\n" % repr(self.batch_size) +
                "\tcorpus=\n%s)" % repr(self.corpus))

    def calc_time(self) -> None:
        """
        Prints statistics about the the total duration of recordings in the
        corpus.
        """

        def get_number_of_frames(feat_fns):
            """ fns: A list of numpy files which contain a number of feature
            frames. """

            total = 0
            for feat_fn in feat_fns:
                num_frames = len(np.load(feat_fn))
                total += num_frames

            return total

        def numframes_to_minutes(num_frames):
            # TODO Assumes 10ms strides for the frames. This should generalize to
            # different frame stride widths, as should feature preparation.
            minutes = ((num_frames*10)/1000)/60
            return minutes

        total_frames = 0

        train_fns = [train_fn[0] for train_fn in self.train_fns]
        num_train_frames = get_number_of_frames(train_fns)
        total_frames += num_train_frames
        num_valid_frames = get_number_of_frames(self.corpus.get_valid_fns()[0])
        total_frames += num_valid_frames
        num_test_frames = get_number_of_frames(self.corpus.get_test_fns()[0])
        total_frames += num_test_frames

        print("Train duration: %0.3f" % numframes_to_minutes(num_train_frames))
        print("Validation duration: %0.3f" % numframes_to_minutes(num_valid_frames))
        print("Test duration: %0.3f" % numframes_to_minutes(num_test_frames))
        print("Total duration: %0.3f" % numframes_to_minutes(total_frames))



"""
Offers functions for tokenizing utterances into phonemes, characters or
other symbols.
"""

from typing import Callable, Iterable, NamedTuple, Set, Dict

from ..utterance import Utterance

LabelSegmenter = NamedTuple("LabelSegmenter",
                            [("segment_labels", Callable[[Utterance], Utterance]),
                            ("labels", Set[str])])
LabelSegmenter.__doc__ = (
    """ An immutable object that segments the phonemes of an utterance. This
    could probably actually have a __call__ implementation. That won't work
    because namedtuples can't have special methods. Perhaps it could instead
    just be a function which we give a labels attribute. Perhaps that
    obfuscates things a bit, but it could be okay.

    Attributes:
        segment_labels: A function that takes an Utterance and returns another
            Utterance where the text field has changed to be phonemically
            segmented, using spaces as delimiters. Eg "this is" -> "th i s i s".
        labels: A set of labels (eg. phonemes or tones) relevant for segmenting.
    """)


def segment_into_chars(utterance: str) -> str:
    """ Segments an utterance into space delimited characters. """

    if not isinstance(utterance, str):
        raise TypeError("Input type must be a string. Got {}.".format(type(utterance)))

    utterance.strip()
    utterance = utterance.replace(" ", "")
    return " ".join(utterance)

def segment_into_tokens(utterance: str, token_inventory: Iterable[str]):
    """
    Segments an utterance (a string) into tokens based on an inventory of
    tokens (a list or set of strings).

    The approach: Given the rest of the utterance, find the largest token (in
    character length) that is found in the token_inventory, and treat that as a
    token before segmenting the rest of the string.

    Note: Your orthography may open the door to ambiguities in the
    segmentation. Hopefully not, but another alternative it to simply segment
    on characters with segment_into_chars()
    """

    if not isinstance(utterance, str):
        raise TypeError("Input type must be a string. Got {}.".format(type(utterance)))

    # Token inventory needs to be hashable for speed
    token_inventory = set(token_inventory)
    # Get the size of the longest token in the inventory
    max_len = len(sorted(list(token_inventory), key=lambda x: len(x))[-1])

    def segment_token(utterance):
        if utterance == "":
            return "", ""

        for i in range(max_len, 0, -1):
            if utterance[:i] in token_inventory:
                return utterance[:i], utterance[i:]
        # If the next character is preventing segmentation, move on.
        # TODO This needs to be logged with a warning on the first occurrence.
        return "", utterance[1:]

    tokens = []
    head, tail = segment_token(utterance)
    tokens.append(head)
    while tail != "":
        head, tail = segment_token(tail)
        tokens.append(head)
    tokens = [tok for tok in tokens if tok != ""]

    return " ".join(tokens)

def make_indices_to_labels(labels: Set[str]) -> Dict[int, str]:
    """ Creates a mapping from indices to labels. """

    return {index: label for index, label in
            enumerate(["pad"] + sorted(list(labels)))}



""" Miscellaneous functions for experiment management. """

import os

import git
from git import Repo

from typing import Optional

import persephone
from . import config
from . import rnn_ctc
from .corpus_reader import CorpusReader
from .utils import is_git_directory_clean

EXP_DIR = config.EXP_DIR # type: str

def get_exp_dir_num(parent_dir: str) -> int:
    """ Gets the number of the current experiment directory."""
    return max([int(fn.split(".")[0])
                for fn in os.listdir(parent_dir) if fn.split(".")[0].isdigit()]
                    + [-1])

def _prepare_directory(directory_path: str) -> str:
    """
    Prepare the directory structure required for the experiment
    :returns: returns the name of the newly created directory
    """
    exp_num = get_exp_dir_num(directory_path)
    exp_num = exp_num + 1
    exp_dir = os.path.join(directory_path, str(exp_num))
    if not os.path.isdir(exp_dir):
        os.makedirs(exp_dir)
    return exp_dir

def prep_sub_exp_dir(parent_dir: str) -> str:
    """ Prepares an experiment subdirectory
    :parent_dir: the parent directory
    :returns: returns the name of the newly created subdirectory
    """
    return _prepare_directory(parent_dir)

def prep_exp_dir(directory=EXP_DIR):
    """ Prepares an experiment directory by copying the code in this directory
    to it as is, and setting the logger to write to files in that directory.
    Copies a git hash of the most changes at git HEAD into the directory to
    keep the experiment results in sync with the version control system.
    :directory: The path to directory we are preparing for the experiment,
                which will be created if it does not currently exist.
    :returns: The name of the newly created experiment directory.
    """
    if not os.path.isdir(directory):
        os.makedirs(directory)
    exp_dir = _prepare_directory(directory)
    try:
        # Get the directory this file is in, so we can grab the git repo.
        dirname = os.path.dirname(os.path.realpath(__file__))
        repo = Repo(dirname, search_parent_directories=True)
        with open(os.path.join(exp_dir, "git_hash.txt"), "w") as f:
            print("SHA1 hash: {hexsha}".format(hexsha=repo.head.commit.hexsha), file=f)
    except git.exc.InvalidGitRepositoryError: # pylint: disable=no-member
        # Then the package was probably installed via pypi. Get the version
        # number instead.
        with open(os.path.join(exp_dir, "version.txt"), "w") as f:
            print("Persephone version {}".format(persephone.__version__), file=f)

    return exp_dir

def get_simple_model(exp_dir, corpus):
    num_layers = 2
    hidden_size= 250

    def decide_batch_size(num_train):

        if num_train >= 512:
            batch_size = 16
        elif num_train < 128:
            if num_train < 4:
                batch_size = 1
            else:
                batch_size = 4
        else:
            batch_size = int(num_train / 32)

        return batch_size

    batch_size = decide_batch_size(len(corpus.train_prefixes))

    corpus_reader = CorpusReader(corpus, batch_size=batch_size)
    model = rnn_ctc.Model(exp_dir, corpus_reader,
                          num_layers=num_layers,
                          hidden_size=hidden_size,
                          decoding_merge_repeated=True)

    return model

def train_ready(corpus, directory=EXP_DIR):

    print(directory)

    exp_dir = prep_exp_dir(directory=directory)
    model = get_simple_model(exp_dir, corpus)
    model.train(min_epochs=20, early_stopping_steps=3)
    return exp_dir

def transcribe(model_path, corpus):
    """ Applies a trained model to untranscribed data in a Corpus. """

    exp_dir = prep_exp_dir()
    model = get_simple_model(exp_dir, corpus)
    model.transcribe(model_path)



""" Generic model for automatic speech recognition. """

import inspect
import itertools
import logging
import math
import os
from pathlib import Path
import sys
from typing import Callable, Optional, Union, Sequence, Set, List, Dict

import tensorflow as tf

from .preprocess import labels, feat_extract
from . import utils
from . import config
from .config import ENCODING
from .corpus import Corpus
from .exceptions import PersephoneException
from .corpus_reader import CorpusReader

allow_growth_config = tf.ConfigProto(log_device_placement=False)
allow_growth_config.gpu_options.allow_growth = True #pylint: disable=no-member

logger = logging.getLogger(__name__) # type: ignore

def load_metagraph(model_path_prefix: Union[str, Path]) -> tf.train.Saver:
    """ Given the path to a model on disk (these will typically be found in
    directories such as exp/<exp_num>/model/model_best.*) creates a Saver
    object that can then be used to restore the graph inside a tf.Session.
    """

    model_path_prefix = str(model_path_prefix)
    metagraph = tf.train.import_meta_graph(model_path_prefix + ".meta")
    return metagraph

def dense_to_human_readable(dense_repr: Sequence[Sequence[int]], index_to_label: Dict[int, str]) -> List[List[str]]:
    """ Converts a dense representation of model decoded output into human
    readable, using a mapping from indices to labels. """

    transcripts = []
    for dense_r in dense_repr:
        non_empty_phonemes = [phn_i for phn_i in dense_r if phn_i != 0]
        transcript = [index_to_label[index] for index in non_empty_phonemes]
        transcripts.append(transcript)

    return transcripts

def decode_corpus(model_path_prefix: Union[str, Path],
                  corpus: Corpus,
                  *,
                  batch_size: int = 64,
                  feat_dir: Optional[Path]=None,
                  batch_x_name: str="batch_x:0",
                  batch_x_lens_name: str="batch_x_lens:0",
                  output_name: str="hyp_dense_decoded:0") -> List[List[str]]:
    input_paths = [Path(corpus.tgt_dir) / "wav" / Path(prefix + ".wav")
                   for prefix in corpus.untranscribed_prefixes]
    return decode(model_path_prefix,
           input_paths,
           label_set=corpus.labels,
           feature_type=corpus.feat_type,
           batch_size=batch_size,
           feat_dir=feat_dir,
           batch_x_name=batch_x_name,
           batch_x_lens_name=batch_x_lens_name,
           output_name=output_name)

def decode(model_path_prefix: Union[str, Path],
           input_paths: Sequence[Path],
           label_set: Set[str],
           *,
           feature_type: str = "fbank", #TODO Make this None and infer feature_type from dimension of NN input layer.
           batch_size: int = 64,
           feat_dir: Optional[Path]=None,
           batch_x_name: str="batch_x:0",
           batch_x_lens_name: str="batch_x_lens:0",
           output_name: str="hyp_dense_decoded:0") -> List[List[str]]:
    """Use an existing tensorflow model that exists on disk to decode
    WAV files.

    Args:
        model_path_prefix: The path to the saved tensorflow model.
                           This is the full prefix to the ".ckpt" file.
        input_paths: A sequence of `pathlib.Path`s to WAV files to put through
                     the model provided.
        label_set: The set of all the labels this model uses.
        feature_type: The type of features this model uses.
                      Note that this MUST match the type of features that the
                      model was trained on initially.
        feat_dir: Any files that require preprocessing will be
                                  saved to the path specified by this.
        batch_x_name: The name of the tensorflow input for batch_x
        batch_x_lens_name: The name of the tensorflow input for batch_x_lens
        output_name: The name of the tensorflow output
    """

    if not input_paths:
        raise PersephoneException("No untranscribed WAVs to transcribe.")

    model_path_prefix = str(model_path_prefix)

    for p in input_paths:
        if not p.exists():
            raise PersephoneException(
                "The WAV file path {} does not exist".format(p)
            )

    preprocessed_file_paths = []
    for p in input_paths:
        prefix = p.stem
        # Check the "feat" directory as per the filesystem conventions of a Corpus
        feature_file_ext = ".{}.npy".format(feature_type)
        conventional_npy_location =  p.parent.parent / "feat" / (Path(prefix + feature_file_ext))
        if conventional_npy_location.exists():
            # don't need to preprocess it
            preprocessed_file_paths.append(conventional_npy_location)
        else:
            if not feat_dir:
                feat_dir = p.parent.parent / "feat"
            if not feat_dir.is_dir():
                os.makedirs(str(feat_dir))

            mono16k_wav_path = feat_dir / "{}.wav".format(prefix)
            feat_path = feat_dir / "{}.{}.npy".format(prefix, feature_type)
            feat_extract.convert_wav(p, mono16k_wav_path)
            preprocessed_file_paths.append(feat_path)
    # preprocess the file that weren't found in the features directory
    # as per the filesystem conventions
    if feat_dir:
        feat_extract.from_dir(feat_dir, feature_type)

    fn_batches = utils.make_batches(preprocessed_file_paths, batch_size)
    # Load the model and perform decoding.
    metagraph = load_metagraph(model_path_prefix)
    with tf.Session() as sess:
        metagraph.restore(sess, model_path_prefix)

        for fn_batch in fn_batches:
            batch_x, batch_x_lens = utils.load_batch_x(fn_batch)

        # TODO These placeholder names should be a backup if names from a newer
        # naming scheme aren't present. Otherwise this won't generalize to
        # different architectures.
        feed_dict = {batch_x_name: batch_x,
                     batch_x_lens_name: batch_x_lens}

        dense_decoded = sess.run(output_name, feed_dict=feed_dict)

    # Create a human-readable representation of the decoded.
    indices_to_labels = labels.make_indices_to_labels(label_set)
    human_readable = dense_to_human_readable(dense_decoded, indices_to_labels)

    return human_readable

class Model:
    """ Generic model for our ASR tasks.

    Attributes:
        exp_dir: Path that the experiment directory is located
        corpus_reader: `CorpusReader` object that provides access to the corpus
                       this model is being trained on.
        log_softmax: log softmax function
        batch_x: A batch of input features. ("x" is the typical notation in ML
                 papers on this topic denoting model input)
        batch_x_lens: The lengths of each utterance. This is used by Tensorflow
                      to know how much to pad utterances that are shorter than
                      this length.
        batch_y: Reference labels for a batch ("y" is the typical notation in ML
                 papers on this topic denoting training labels)
        optimizer: The gradient descent method being used. (Typically we use Adam
                   because it has provided good results but any stochastic gradient
                   descent method could be substituted here)
        ler: Label error rate.
        dense_decoded: Dense representation of the model transcription output.
        dense_ref: Dense representation of the reference transcription.
        saved_model_path: Path to where the Tensorflow model is being saved on disk.
    """

    def __init__(self, exp_dir: Union[Path, str], corpus_reader: CorpusReader) -> None:
        self.exp_dir = str(exp_dir) if isinstance(exp_dir, Path) else exp_dir # type: str
        self.corpus_reader = corpus_reader
        self.log_softmax = None
        self.batch_x = None
        self.batch_x_lens = None
        self.batch_y = None
        self.optimizer = None
        self.ler = None
        self.dense_decoded = None
        self.dense_ref = None
        self.saved_model_path = "" # type: str

    def transcribe(self, restore_model_path: Optional[str]=None) -> None:
        """ Transcribes an untranscribed dataset. Similar to eval() except
        no reference translation is assumed, thus no LER is calculated.
        """

        saver = tf.train.Saver()
        with tf.Session(config=allow_growth_config) as sess:
            if restore_model_path:
                saver.restore(sess, restore_model_path)
            else:
                if self.saved_model_path:
                    saver.restore(sess, self.saved_model_path)
                else:
                    raise PersephoneException("No model to use for transcription.")

            batch_gen = self.corpus_reader.untranscribed_batch_gen()

            hyp_batches = []
            for batch_i, batch in enumerate(batch_gen):

                batch_x, batch_x_lens, feat_fn_batch = batch

                feed_dict = {self.batch_x: batch_x,
                             self.batch_x_lens: batch_x_lens}

                [dense_decoded] = sess.run([self.dense_decoded], feed_dict=feed_dict)
                hyps = self.corpus_reader.human_readable(dense_decoded)

                # Prepare dir for transcription
                hyps_dir = os.path.join(self.exp_dir, "transcriptions")
                if not os.path.isdir(hyps_dir):
                    os.mkdir(hyps_dir)

                hyp_batches.append((hyps,feat_fn_batch))

            with open(os.path.join(hyps_dir, "hyps.txt"), "w",
                      encoding=ENCODING) as hyps_f:
                for hyp_batch, fn_batch in hyp_batches:
                    for hyp, fn in zip(hyp_batch, fn_batch):
                        print(fn, file=hyps_f)
                        print(" ".join(hyp), file=hyps_f)
                        print("", file=hyps_f)

    def decode(self):
        model_path_prefix = Path(self.exp_dir) / "model" / "model_best.ckpt"
        prefixes = self.corpus_reader.corpus.untranscribed_prefixes
        input_paths = [self.corpus_reader.corpus.tgt_dir / "feat" / Path(p + ".wav")
                       for p in prefixes]
        label_set = self.corpus_reader.corpus.labels
        feature_type = self.corpus_reader.corpus.feat_type
        batch_size = self.corpus_reader.batch_size
        batch_x_name = self.batch_x.name
        batch_x_lens_name = self.batch_x_lens.name
        output_name = self.dense_decoded.name
        return decode(model_path_prefix,
               input_paths,
               label_set,
               feature_type=feature_type,
               batch_size=batch_size,
               batch_x_name=batch_x_name,
               batch_x_lens_name=batch_x_lens_name,
               output_name=output_name)
 
    def eval(self, restore_model_path: Optional[str]=None) -> None:
        """ Evaluates the model on a test set."""

        saver = tf.train.Saver()
        with tf.Session(config=allow_growth_config) as sess:
            if restore_model_path:
                logger.info("restoring model from %s", restore_model_path)
                saver.restore(sess, restore_model_path)
            else:
                assert self.saved_model_path, "{}".format(self.saved_model_path)
                logger.info("restoring model from %s", self.saved_model_path)
                saver.restore(sess, self.saved_model_path)

            test_x, test_x_lens, test_y = self.corpus_reader.test_batch()

            feed_dict = {self.batch_x: test_x,
                         self.batch_x_lens: test_x_lens,
                         self.batch_y: test_y}

            test_ler, dense_decoded, dense_ref = sess.run(
                [self.ler, self.dense_decoded, self.dense_ref],
                feed_dict=feed_dict)
            hyps, refs = self.corpus_reader.human_readable_hyp_ref(
                dense_decoded, dense_ref)
            # Log hypotheses
            hyps_dir = os.path.join(self.exp_dir, "test")
            if not os.path.isdir(hyps_dir):
                os.mkdir(hyps_dir)
            with open(os.path.join(hyps_dir, "hyps"), "w",
                      encoding=ENCODING) as hyps_f:
                for hyp in hyps:
                    print(" ".join(hyp), file=hyps_f)
            with open(os.path.join(hyps_dir, "refs"), "w",
                      encoding=ENCODING) as refs_f:
                for ref in refs:
                    print(" ".join(ref), file=refs_f)

            test_per = utils.batch_per(hyps, refs)
            if not math.isclose(test_per, test_ler, rel_tol=1e-07):
                logger.warning("The label error rate from Tensorflow doesn't exactly"
                                "match the phoneme error rate calculated in persephone"
                                "Tensorflow %f, Persephone %f", test_ler, test_per)
            with open(os.path.join(hyps_dir, "test_per"), "w",
                      encoding=ENCODING) as per_f:
                print("LER: %f" % (test_ler), file=per_f)

    def output_best_scores(self, best_epoch_str: str) -> None:
        """Output best scores to the filesystem"""
        BEST_SCORES_FILENAME = "best_scores.txt"
        with open(os.path.join(self.exp_dir, BEST_SCORES_FILENAME),
                  "w", encoding=ENCODING) as best_f:
            print(best_epoch_str, file=best_f, flush=True)

    def train(self, *, early_stopping_steps: int = 10, min_epochs: int = 30,
              max_valid_ler: float = 1.0, max_train_ler: float = 0.3,
              max_epochs: int = 100, restore_model_path: Optional[str]=None,
              epoch_callback: Optional[Callable[[Dict], None]]=None) -> None:
        """ Train the model.

            min_epochs: minimum number of epochs to run training for.
            max_epochs: maximum number of epochs to run training for.
            early_stopping_steps: Stop training after this number of steps
                                  if no LER improvement has been made.
            max_valid_ler: Maximum LER for the validation set.
                           Training will continue until this is met or another
                           stopping condition occurs.
            max_train_ler: Maximum LER for the training set.
                           Training will continue until this is met or another
                           stopping condition occurs.
            restore_model_path: The path to restore a model from.
            epoch_callback: A callback that is called at the end of each training epoch.
                            The parameters passed to the callable will be the epoch number,
                            the current training LER and the current validation LER.
                            This can be useful for progress reporting.
        """
        logger.info("Training model")
        best_valid_ler = 2.0
        steps_since_last_record = 0

        #Get information about training for the names of output files.
        frame = inspect.currentframe()
        # pylint: disable=deprecated-method
        # It was a mistake to deprecate this in Python 3.5
        if frame:
            args, _, _, values = inspect.getargvalues(frame)
            with open(os.path.join(self.exp_dir, "train_description.txt"), 
                      "w", encoding=ENCODING) as desc_f:
                for arg in args:
                    if type(values[arg]) in [str, int, float] or isinstance(
                            values[arg], type(None)):
                        print("%s=%s" % (arg, values[arg]), file=desc_f)
                    else:
                        print("%s=%s" % (arg, values[arg].__dict__), file=desc_f)
                print("num_train=%s" % (self.corpus_reader.num_train), file=desc_f)
                print("batch_size=%s" % (self.corpus_reader.batch_size), file=desc_f)
        else:
            logger.error("Couldn't find frame information, failed to write train_description.txt")


        # Load the validation set
        valid_x, valid_x_lens, valid_y = self.corpus_reader.valid_batch()

        saver = tf.train.Saver()

        with tf.Session(config=allow_growth_config) as sess:

            if restore_model_path:
                logger.info("Restoring model from path %s", restore_model_path)
                saver.restore(sess, restore_model_path)
            else:
                sess.run(tf.global_variables_initializer())

            # Prepare directory to output hypotheses to
            hyps_dir = os.path.join(self.exp_dir, "decoded")
            if not os.path.isdir(hyps_dir):
                os.mkdir(hyps_dir)

            best_epoch_str = None

            training_log_path = os.path.join(self.exp_dir, "train_log.txt")
            if os.path.exists(training_log_path):
                logger.error("Error, overwriting existing log file at path {}".format(training_log_path))
            with open(training_log_path, "w",
                      encoding=ENCODING) as out_file:
                for epoch in itertools.count(start=1):
                    print("\nexp_dir %s, epoch %d" % (self.exp_dir, epoch))
                    batch_gen = self.corpus_reader.train_batch_gen()

                    train_ler_total = 0
                    print("\tBatch...", end="")
                    for batch_i, batch in enumerate(batch_gen):
                        print("%d..." % batch_i, end="")
                        sys.stdout.flush()
                        batch_x, batch_x_lens, batch_y = batch

                        feed_dict = {self.batch_x: batch_x,
                                    self.batch_x_lens: batch_x_lens,
                                    self.batch_y: batch_y}

                        _, ler, = sess.run([self.optimizer, self.ler],
                                        feed_dict=feed_dict)

                        train_ler_total += ler
                    #else:
                    #    raise PersephoneException("No training data was provided."
                    #                              " Check your batch generation.")

                    feed_dict = {self.batch_x: valid_x,
                                self.batch_x_lens: valid_x_lens,
                                self.batch_y: valid_y}

                    try:
                        valid_ler, dense_decoded, dense_ref = sess.run(
                            [self.ler, self.dense_decoded, self.dense_ref],
                            feed_dict=feed_dict)
                    except tf.errors.ResourceExhaustedError:
                        import pprint
                        print("Ran out of memory allocating a batch:")
                        pprint.pprint(feed_dict)
                        logger.critical("Ran out of memory allocating a batch: %s", pprint.pformat(feed_dict))
                        raise
                    hyps, refs = self.corpus_reader.human_readable_hyp_ref(
                        dense_decoded, dense_ref)
                    # Log hypotheses
                    with open(os.path.join(hyps_dir, "epoch%d_hyps" % epoch),
                              "w", encoding=ENCODING) as hyps_f:
                        for hyp in hyps:
                            print(" ".join(hyp), file=hyps_f)
                    if epoch == 1:
                        with open(os.path.join(hyps_dir, "refs"), "w", 
                                  encoding=ENCODING) as refs_f:
                            for ref in refs:
                                print(" ".join(ref), file=refs_f)

                    valid_per = utils.batch_per(hyps, refs)

                    epoch_str = "Epoch %d. Training LER: %f, validation LER: %f" % (
                        epoch, (train_ler_total / (batch_i + 1)), valid_ler)
                    print(epoch_str, flush=True, file=out_file)
                    if best_epoch_str is None:
                        best_epoch_str = epoch_str

                    # Call the callback here if it was defined
                    if epoch_callback:
                        epoch_callback({
                            "epoch": epoch,
                            "training_ler": (train_ler_total / (batch_i + 1)), # current training LER
                            "valid_ler": valid_ler, # Current validation LER
                        })

                    # Implement early stopping.
                    if valid_ler < best_valid_ler:
                        print("New best valid_ler", file=out_file)
                        best_valid_ler = valid_ler
                        best_epoch_str = epoch_str
                        steps_since_last_record = 0

                        # Save the model.
                        checkpoint_path = os.path.join(self.exp_dir, "model", "model_best.ckpt")
                        if not os.path.exists(os.path.dirname(checkpoint_path)):
                            os.mkdir(os.path.dirname(checkpoint_path))
                        saver.save(sess, checkpoint_path)
                        self.saved_model_path = checkpoint_path

                        # Output best hyps
                        with open(os.path.join(hyps_dir, "best_hyps"),
                                  "w", encoding=ENCODING) as hyps_f:
                            for hyp in hyps:
                                print(" ".join(hyp), file=hyps_f)

                    else:
                        print("Steps since last best valid_ler: %d" % (steps_since_last_record), file=out_file)
                        steps_since_last_record += 1
                        if epoch >= max_epochs:
                            self.output_best_scores(best_epoch_str)
                            break
                        if steps_since_last_record >= early_stopping_steps:
                            if epoch >= min_epochs:
                                # Then we've done the minimum number of epochs.
                                if valid_ler <= max_valid_ler and ler <= max_train_ler:
                                    # Then training error has moved sufficiently
                                    # towards convergence.
                                    print("Stopping since best validation score hasn't been"
                                        " beaten in %d epochs and at least %d have been"
                                        " done. The valid ler (%d) is below %d and"
                                        " the train ler (%d) is below %d." %
                                        (early_stopping_steps, min_epochs, valid_ler,
                                        max_valid_ler, ler, max_train_ler),
                                        file=out_file, flush=True)
                                    self.output_best_scores(best_epoch_str)
                                    break
                                else:
                                    # Keep training because we haven't achieved
                                    # convergence.
                                    continue
                            else:
                                # Keep training because we haven't done the minimum
                                # numper of epochs.
                                continue

                # Check we actually saved a checkpoint
                if not self.saved_model_path:
                    raise PersephoneException(
                        "No checkpoint was saved so model evaluation cannot be performed. "
                        "This can happen if the validaion LER never converges.")
                # Finally, run evaluation on the test set.
                self.eval(restore_model_path=self.saved_model_path)



""" An interface with the Na data. """

import logging
from pathlib import Path
import os
import random

import pint # type: ignore

from .. import config
from .. import corpus
from ..preprocess import feat_extract
from ..preprocess import wav
from .. import utils
from ..exceptions import PersephoneException
from ..preprocess import pangloss

ureg = pint.UnitRegistry()

logging.config.fileConfig(config.LOGGING_INI_PATH)

ORG_DIR = config.NA_PATH
TGT_DIR = os.path.join(config.TGT_DIR, "na")
ORG_XML_DIR = os.path.join(ORG_DIR, "xml/TEXT/F4/")
ORG_WAV_DIR = os.path.join(ORG_DIR, "wav")
TGT_WAV_DIR = os.path.join(TGT_DIR, "wav")
FEAT_DIR = os.path.join(TGT_DIR, "feat")
LABEL_DIR = os.path.join(TGT_DIR, "label")
TRANSL_DIR = os.path.join(TGT_DIR, "transl")

#PREFIXES = [os.path.splitext(fn)[0]
#            for fn in os.listdir(ORG_TRANSCRIPT_DIR)
#            if fn.endswith(".txt")]

MISC_SYMBOLS = [' ̩', '~', '=', ':', 'F', '¨', '↑', '“', '”', '…', '«', '»',
'D', 'a', 'ː', '#', '$', "‡"]
BAD_NA_SYMBOLS = ['D', 'F', '~', '…', '=', '↑', ':']
PUNC_SYMBOLS = [',', '!', '.', ';', '?', "'", '"', '*', ':', '«', '»', '“', '”', "ʔ"]
UNI_PHNS = {'q', 'p', 'ɭ', 'ɳ', 'h', 'ʐ', 'n', 'o', 'ɤ', 'ʝ', 'ɛ', 'g',
            'i', 'u', 'b', 'ɔ', 'ɯ', 'v', 'ɑ', 'l', 'ɖ', 'ɻ', 'ĩ', 'm',
            't', 'w', 'õ', 'ẽ', 'd', 'ɣ', 'ɕ', 'c', 'ʁ', 'ʑ', 'ʈ', 'ɲ', 'ɬ',
            's', 'ŋ', 'ə', 'e', 'æ', 'f', 'j', 'k', 'z', 'ʂ'}
BI_PHNS = {'dʑ', 'ẽ', 'ɖʐ', 'w̃', 'æ̃', 'qʰ', 'i͂', 'tɕ', 'v̩', 'o̥', 'ts',
           'ɻ̩', 'ã', 'ə̃', 'ṽ', 'pʰ', 'tʰ', 'ɤ̃', 'ʈʰ', 'ʈʂ', 'ɑ̃', 'ɻ̃', 'kʰ',
           'ĩ', 'õ', 'dz', "ɻ̍", "wæ", "wɑ", "wɤ", "jæ", "jɤ", "jo"}
FILLERS = {"əəə…", "mmm…"}
TRI_PHNS = {"tɕʰ", "ʈʂʰ", "tsʰ", "ṽ̩", "ṽ̩", "ɻ̩̃", "wæ̃", "w̃æ"}
UNI_TONES = {"˩", "˥", "˧"}
BI_TONES = {"˧˥", "˩˥", "˩˧", "˧˩"}
TONES = UNI_TONES.union(BI_TONES)
SYMBOLS_TO_PREDICT = {"|"}

PHONEMES = UNI_PHNS.union(BI_PHNS).union(TRI_PHNS).union(FILLERS)

# TODO Get rid of these variables, as they're not used in the class, only for
# preparing phonemes_onehot feats.
PHONEMES_TO_INDICES = {phn: index for index, phn in enumerate(PHONEMES)}
INDICES_TO_PHONEMES = {index: phn for index, phn in enumerate(PHONEMES)}

# TODO Potentially remove?
#PHONES_TONES = sorted(list(PHONES.union(set(TONES)))) # Sort for determinism
#PHONESTONES2INDICES = {phn_tone: index for index, phn_tone in enumerate(PHONES_TONES)}
#INDICES2PHONESTONES = {index: phn_tone for index, phn_tone in enumerate(PHONES_TONES)}
#TONES2INDICES = {tone: index for index, tone in enumerate(TONES)}
#INDICES2TONES = {index: tone for index, tone in enumerate(TONES)}

def preprocess_na(sent, label_type):
    """Preprocess Na sentences

    Args:
        sent: A sentence
        label_type: The type of label provided
    """
    if label_type == "phonemes_and_tones":
        phonemes = True
        tones = True
        tgm = True
    elif label_type == "phonemes_and_tones_no_tgm":
        phonemes = True
        tones = True
        tgm = False
    elif label_type == "phonemes":
        phonemes = True
        tones = False
        tgm = False
    elif label_type == "tones":
        phonemes = False
        tones = True
        tgm = True
    elif label_type == "tones_notgm":
        phonemes = False
        tones = True
        tgm = False
    else:
        raise ValueError("Unrecognized label type: %s" % label_type)

    def pop_phoneme(sentence):
        """Pop phonemes off a sentence one at a time"""
        # TODO desperately needs refactoring

        # Treating fillers as single tokens; normalizing to əəə and mmm
        if phonemes:
            if sentence[:4] in ["əəə…", "mmm…"]:
                return sentence[:4], sentence[4:]
            if sentence.startswith("ə…"):
                return "əəə…", sentence[2:]
            if sentence.startswith("m…"):
                return "mmm…", sentence[2:]
            if sentence.startswith("mm…"):
                return "mmm…", sentence[3:]

        # Normalizing some stuff
        if sentence[:3] == "wæ̃":
            if phonemes:
                return "w̃æ", sentence[3:]
            else:
                return None, sentence[3:]
        if sentence[:3] == "ṽ̩":
            if phonemes:
                return "ṽ̩", sentence[3:]
            else:
                return None, sentence[3:]

        if sentence[:3] in TRI_PHNS:
            if phonemes:
                return sentence[:3], sentence[3:]
            else:
                return None, sentence[3:]
        if sentence[:2] in BI_PHNS:
            if phonemes:
                return sentence[:2], sentence[2:]
            else:
                return None, sentence[2:]
        if sentence[:2] == "˧̩":
            return "˧", sentence[2:]
        if sentence[:2] == "˧̍":
            return "˧", sentence[2:]
        if sentence[0] in UNI_PHNS:
            if phonemes:
                return sentence[0], sentence[1:]
            else:
                return None, sentence[1:]
        if sentence[:2] in BI_TONES:
            if tones:
                return sentence[:2], sentence[2:]
            else:
                return None, sentence[2:]
        if sentence[0] in UNI_TONES:
            if tones:
                return sentence[0], sentence[1:]
            else:
                return None, sentence[1:]
        if sentence[0] in MISC_SYMBOLS:
            # We assume these symbols cannot be captured.
            return None, sentence[1:]
        if sentence[0] in BAD_NA_SYMBOLS:
            return None, sentence[1:]
        if sentence[0] in PUNC_SYMBOLS:
            return None, sentence[1:]
        if sentence[0] in ["-", "ʰ", "/"]:
            return None, sentence[1:]
        if sentence[0] in set(["<", ">"]):
            # We keep everything literal, thus including what is in <>
            # brackets; so we just remove these tokens"
            return None, sentence[1:]
        if sentence[0] == "[":
            # It's an opening square bracket, so ignore everything until we
            # find a closing one.
            if sentence.find("]") == len(sentence)-1:
                # If the closing bracket is the last char
                return None, ""
            else:
                return None, sentence[sentence.find("]")+1:]
        if sentence[0] in set([" ", "\t", "\n"]):
            # Return a space char so that it can be identified in word segmentation
            # processing.
            return " ", sentence[1:]
        if sentence[0] == "|" or sentence[0] == "ǀ" or sentence[0] == "◊":
            # TODO Address extrametrical span symbol ◊ differently. For now,
            # treating it as a tone group boundary marker for consistency with
            # previous work.
            if tgm:
                return "|", sentence[1:]
            else:
                return None, sentence[1:]
        if sentence[0] in "()":
            return None, sentence[1:]
        print("***" + sentence)
        raise ValueError("Next character not recognized: " + sentence[:1])

    def filter_for_phonemes(sentence):
        """ Returns a sequence of phonemes and pipes (word delimiters). Tones,
        syllable boundaries, whitespace are all removed."""

        filtered_sentence = []
        while sentence != "":
            phoneme, sentence = pop_phoneme(sentence)
            if phoneme != " ":
                filtered_sentence.append(phoneme)
        filtered_sentence = [item for item in filtered_sentence if item != None]
        return " ".join(filtered_sentence)

    # Filter utterances with certain words
    if "BEGAIEMENT" in sent:
        return ""
    sent = filter_for_phonemes(sent)
    return sent

def preprocess_french(trans, fr_nlp, remove_brackets_content=True):
    """ Takes a list of sentences in french and preprocesses them."""

    if remove_brackets_content:
        trans = pangloss.remove_content_in_brackets(trans, "[]")
    # Not sure why I have to split and rejoin, but that fixes a Spacy token
    # error.
    trans = fr_nlp(" ".join(trans.split()[:]))
    #trans = fr_nlp(trans)
    trans = " ".join([token.lower_ for token in trans if not token.is_punct])

    return trans

def trim_wavs(org_wav_dir=ORG_WAV_DIR,
              tgt_wav_dir=TGT_WAV_DIR,
              org_xml_dir=ORG_XML_DIR):
    """ Extracts sentence-level transcriptions, translations and wavs from the
    Na Pangloss XML and WAV files. But otherwise doesn't preprocess them."""

    logging.info("Trimming wavs...")

    if not os.path.exists(os.path.join(tgt_wav_dir, "TEXT")):
        os.makedirs(os.path.join(tgt_wav_dir, "TEXT"))
    if not os.path.exists(os.path.join(tgt_wav_dir, "WORDLIST")):
        os.makedirs(os.path.join(tgt_wav_dir, "WORDLIST"))

    for fn in os.listdir(org_xml_dir):
        path = os.path.join(org_xml_dir, fn)
        prefix, _ = os.path.splitext(fn)

        if os.path.isdir(path):
            continue
        if not path.endswith(".xml"):
            continue

        logging.info("Trimming wavs from {}".format(fn))

        rec_type, _, times, _ = pangloss.get_sents_times_and_translations(path)

        # Extract the wavs given the times.
        for i, (start_time, end_time) in enumerate(times):
            if prefix.endswith("PLUSEGG"):
                in_wav_path = os.path.join(org_wav_dir, prefix.upper()[:-len("PLUSEGG")]) + ".wav"
            else:
                in_wav_path = os.path.join(org_wav_dir, prefix.upper()) + ".wav"
            headmic_path = os.path.join(org_wav_dir, prefix.upper()) + "_HEADMIC.wav"
            if os.path.isfile(headmic_path):
                in_wav_path = headmic_path

            out_wav_path = os.path.join(tgt_wav_dir, rec_type, "%s.%d.wav" % (prefix, i))
            if not os.path.isfile(in_wav_path):
                raise PersephoneException("{} not a file.".format(in_wav_path))
            start_time = start_time * ureg.seconds
            end_time = end_time * ureg.seconds
            wav.trim_wav_ms(Path(in_wav_path), Path(out_wav_path),
                            start_time.to(ureg.milliseconds).magnitude,
                            end_time.to(ureg.milliseconds).magnitude)

def prepare_labels(label_type, org_xml_dir=ORG_XML_DIR, label_dir=LABEL_DIR):
    """ Prepare the neural network output targets."""

    if not os.path.exists(os.path.join(label_dir, "TEXT")):
        os.makedirs(os.path.join(label_dir, "TEXT"))
    if not os.path.exists(os.path.join(label_dir, "WORDLIST")):
        os.makedirs(os.path.join(label_dir, "WORDLIST"))

    for path in Path(org_xml_dir).glob("*.xml"):
        fn = path.name
        prefix, _ = os.path.splitext(fn)

        rec_type, sents, _, _ = pangloss.get_sents_times_and_translations(str(path))
        # Write the sentence transcriptions to file
        sents = [preprocess_na(sent, label_type) for sent in sents]
        for i, sent in enumerate(sents):
            if sent.strip() == "":
                # Then there's no transcription, so ignore this.
                continue
            out_fn = "%s.%d.%s" % (prefix, i, label_type)
            sent_path = os.path.join(label_dir, rec_type, out_fn)
            with open(sent_path, "w") as sent_f:
                print(sent, file=sent_f)

# TODO Consider factoring out as non-Na specific.
def prepare_untran(feat_type, tgt_dir, untran_dir):
    """ Preprocesses untranscribed audio."""
    org_dir = str(untran_dir)
    wav_dir = os.path.join(str(tgt_dir), "wav", "untranscribed")
    feat_dir = os.path.join(str(tgt_dir), "feat", "untranscribed")
    if not os.path.isdir(wav_dir):
        os.makedirs(wav_dir)
    if not os.path.isdir(feat_dir):
        os.makedirs(feat_dir)

    # Standardize into wav files
    for fn in os.listdir(org_dir):
        in_path = os.path.join(org_dir, fn)
        prefix, _ = os.path.splitext(fn)
        mono16k_wav_path = os.path.join(wav_dir, "%s.wav" % prefix)
        if not os.path.isfile(mono16k_wav_path):
            feat_extract.convert_wav(Path(in_path), Path(mono16k_wav_path))

    # Split up the wavs and write prefixes to prefix file.
    wav_fns = os.listdir(wav_dir)
    with (tgt_dir / "untranscribed_prefixes.txt").open("w") as prefix_f:
        for fn in wav_fns:
            in_fn = os.path.join(wav_dir, fn)
            prefix, _ = os.path.splitext(fn)
            # Split into sub-wavs and perform feat extraction.
            split_id = 0
            start, end = 0, 10 #in seconds
            length = utils.wav_length(in_fn)
            while True:
                sub_wav_prefix = "{}.{}".format(prefix, split_id)
                print(sub_wav_prefix, file=prefix_f)
                out_fn = os.path.join(feat_dir, "{}.wav".format(sub_wav_prefix))
                start_time = start * ureg.seconds
                end_time = end * ureg.seconds
                if not Path(out_fn).is_file():
                    wav.trim_wav_ms(Path(in_fn), Path(out_fn),
                                    start_time.to(ureg.milliseconds).magnitude,
                                    end_time.to(ureg.milliseconds).magnitude)
                if end > length:
                    break
                start += 10
                end += 10
                split_id += 1

    # Do feat extraction.
    feat_extract.from_dir(Path(os.path.join(feat_dir)), feat_type=feat_type)

# TODO Consider factoring out as non-Na specific
def prepare_feats(feat_type, org_wav_dir=ORG_WAV_DIR, feat_dir=FEAT_DIR, tgt_wav_dir=TGT_WAV_DIR,
                  org_xml_dir=ORG_XML_DIR, label_dir=LABEL_DIR):
    """ Prepare the input features."""

    if not os.path.isdir(TGT_DIR):
        os.makedirs(TGT_DIR)

    if not os.path.isdir(FEAT_DIR):
        os.makedirs(FEAT_DIR)

    if not os.path.isdir(os.path.join(feat_dir, "WORDLIST")):
        os.makedirs(os.path.join(feat_dir, "WORDLIST"))
    if not os.path.isdir(os.path.join(feat_dir, "TEXT")):
        os.makedirs(os.path.join(feat_dir, "TEXT"))

    # Extract utterances from WAVS.
    trim_wavs(org_wav_dir=org_wav_dir,
              tgt_wav_dir=tgt_wav_dir,
              org_xml_dir=org_xml_dir)

    # TODO Currently assumes that the wav trimming from XML has already been
    # done.
    prefixes = []
    for fn in os.listdir(os.path.join(tgt_wav_dir, "WORDLIST")):
        if fn.endswith(".wav"):
            pre, _ = os.path.splitext(fn)
            prefixes.append(os.path.join("WORDLIST", pre))
    for fn in os.listdir(os.path.join(tgt_wav_dir, "TEXT")):
        if fn.endswith(".wav"):
            pre, _ = os.path.splitext(fn)
            prefixes.append(os.path.join("TEXT", pre))

    if feat_type=="phonemes_onehot":
        import numpy as np
        #prepare_labels("phonemes")
        for prefix in prefixes:
            label_fn = os.path.join(label_dir, "%s.phonemes" % prefix)
            out_fn = os.path.join(feat_dir, "%s.phonemes_onehot" %  prefix)
            try:
                with open(label_fn) as label_f:
                    labels = label_f.readlines()[0].split()
            except FileNotFoundError:
                continue
            indices = [PHONEMES_TO_INDICES[label] for label in labels]
            one_hots = [[0]*len(PHONEMES) for _ in labels]
            for i, index in enumerate(indices):
                one_hots[i][index] = 1
                one_hots = np.array(one_hots)
                np.save(out_fn, one_hots)
    else:
        # Otherwise, 
        for prefix in prefixes:
            # Convert the wave to 16k mono.
            wav_fn = os.path.join(tgt_wav_dir, "%s.wav" % prefix)
            mono16k_wav_fn = os.path.join(feat_dir, "%s.wav" % prefix)
            if not os.path.isfile(mono16k_wav_fn):
                logging.info("Normalizing wav {} to a 16k 16KHz mono {}".format(
                    wav_fn, mono16k_wav_fn))
                feat_extract.convert_wav(wav_fn, mono16k_wav_fn)

        # Extract features from the wavs.
        feat_extract.from_dir(Path(os.path.join(feat_dir, "WORDLIST")), feat_type=feat_type)
        feat_extract.from_dir(Path(os.path.join(feat_dir, "TEXT")), feat_type=feat_type)

def get_story_prefixes(label_type, label_dir=LABEL_DIR):
    """ Gets the Na text prefixes. """
    prefixes = [prefix for prefix in os.listdir(os.path.join(label_dir, "TEXT"))
                if prefix.endswith(".%s" % label_type)]
    prefixes = [os.path.splitext(os.path.join("TEXT", prefix))[0]
                for prefix in prefixes]
    return prefixes

def make_data_splits(label_type, train_rec_type="text_and_wordlist", max_samples=1000,
                     seed=0, tgt_dir=TGT_DIR):
    """ Creates a file with a list of prefixes (identifiers) of utterances to
    include in the test set. Test utterances must never be wordlists. Assumes
    preprocessing of label dir has already been done."""

    feat_dir = os.path.join(tgt_dir, "feat")
    test_prefix_fn=os.path.join(tgt_dir, "test_prefixes.txt")
    valid_prefix_fn=os.path.join(tgt_dir, "valid_prefixes.txt")
    with open(test_prefix_fn) as f:
        prefixes = f.readlines()
        test_prefixes = [("TEXT/" + prefix).strip() for prefix in prefixes]
    with open(valid_prefix_fn) as f:
        prefixes = f.readlines()
        valid_prefixes = [("TEXT/" + prefix).strip() for prefix in prefixes]

    label_dir = os.path.join(tgt_dir, "label")
    prefixes = get_story_prefixes(label_type, label_dir=label_dir)
    prefixes = list(set(prefixes) - set(valid_prefixes))
    prefixes = list(set(prefixes) - set(test_prefixes))
    prefixes = utils.filter_by_size(
        feat_dir, prefixes, "fbank", max_samples)

    if train_rec_type == "text":
        train_prefixes = prefixes
    else:
        wordlist_prefixes = [prefix for prefix in os.listdir(os.path.join(label_dir, "WORDLIST"))
                             if prefix.endswith("phonemes")]
        wordlist_prefixes = [os.path.splitext(os.path.join("WORDLIST", prefix))[0]
                             for prefix in wordlist_prefixes]
        wordlist_prefixes = utils.filter_by_size(
                feat_dir, wordlist_prefixes, "fbank", max_samples)
        if train_rec_type == "wordlist":
            prefixes = wordlist_prefixes
        elif train_rec_type == "text_and_wordlist":
            prefixes.extend(wordlist_prefixes)
        else:
            raise PersephoneException("train_rec_type='%s' not supported." % train_rec_type)
        train_prefixes = prefixes
    random.seed(seed)
    random.shuffle(train_prefixes)

    return train_prefixes, valid_prefixes, test_prefixes

def get_stories(label_type):
    """ Returns a list of the stories in the Na corpus. """

    prefixes = get_story_prefixes(label_type)
    texts = list(set([prefix.split(".")[0].split("/")[1] for prefix in prefixes]))
    return texts

def make_story_splits(valid_story, test_story, max_samples, label_type, tgt_dir=TGT_DIR):

    feat_dir=os.path.join(tgt_dir, "feat/")

    prefixes = get_story_prefixes(label_type)
    # TODO Remove assumption of fbank features
    prefixes = utils.filter_by_size(
        feat_dir, prefixes, "fbank", max_samples)

    train = []
    valid = []
    test = []
    for prefix in prefixes:
        if valid_story == os.path.basename(prefix).split(".")[0]:
            valid.append(prefix)
        elif test_story == os.path.basename(prefix).split(".")[0]:
            test.append(prefix)
        else:
            train.append(prefix)

    # Sort by utterance integer value
    test.sort(key=lambda x: int(x.split(".")[-1]))
    valid.sort(key=lambda x: int(x.split(".")[-1]))

    return train, valid, test

class Corpus(corpus.Corpus):
    """ Class to interface with the Na corpus. """

    def __init__(self,
                 feat_type="fbank_and_pitch",
                 label_type="phonemes_and_tones",
                 train_rec_type="text", max_samples=1000,
                 valid_story=None, test_story=None,
                 tgt_dir=Path(TGT_DIR)):

        self.tgt_dir = tgt_dir
        self.get_wav_dir().mkdir(parents=True, exist_ok=True) # pylint: disable=no-member
        self.get_label_dir().mkdir(exist_ok=True) # pylint: disable=no-member

        self.valid_story = valid_story
        self.test_story = test_story

        self.max_samples = max_samples
        self.train_rec_type = train_rec_type

        if label_type == "phonemes_and_tones":
            self.labels = PHONEMES.union(TONES).union(SYMBOLS_TO_PREDICT)
        elif label_type == "phonemes_and_tones_no_tgm":
            self.labels = PHONEMES.union(TONES)
        elif label_type == "phonemes":
            self.labels = PHONEMES
        elif label_type == "tones":
            self.labels = TONES.union(SYMBOLS_TO_PREDICT)
        elif label_type == "tones_notgm":
            self.labels = TONES
        else:
            raise PersephoneException("label_type %s not implemented." % label_type)

        tgt_label_dir = str(tgt_dir / "label")
        tgt_wav_dir = str(tgt_dir / "wav")
        tgt_feat_dir = str(tgt_dir / "feat")
        prepare_labels(label_type, label_dir=tgt_label_dir)
        prepare_feats(feat_type, tgt_wav_dir=tgt_wav_dir,
                                 feat_dir=tgt_feat_dir,
                                 label_dir=tgt_label_dir)
        untran_dir = Path(config.NA_PATH) / "untranscribed_wav"
        logging.debug(untran_dir)
        if untran_dir.is_dir(): # pylint: disable=no-member
            prepare_untran(feat_type=feat_type, tgt_dir=tgt_dir,
                           untran_dir=untran_dir)

        super().__init__(feat_type, label_type, tgt_dir, self.labels, max_samples=max_samples)

    def make_data_splits(self, max_samples, valid_story=None, test_story=None):
        """Split data into train, valid and test groups"""
        # TODO Make this also work with wordlists.
        if valid_story or test_story:
            if not (valid_story and test_story):
                raise PersephoneException(
                    "We need a valid story if we specify a test story "
                    "and vice versa. This shouldn't be required but for "
                    "now it is.")

            train, valid, test = make_story_splits(valid_story, test_story,
                                                   max_samples,
                                                   self.label_type,
                                                   tgt_dir=str(self.tgt_dir))
        else:
            train, valid, test = make_data_splits(self.label_type,
                                                  train_rec_type=self.train_rec_type,
                                                  max_samples=max_samples,
                                                  tgt_dir=str(self.tgt_dir))
        self.train_prefixes = train
        self.valid_prefixes = valid
        self.test_prefixes = test

    def output_story_prefixes(self):
        """ Writes the set of prefixes to a file this is useful for pretty
        printing in results.latex_output. """

        if not self.test_story:
            raise NotImplementedError(
                "I want to write the prefixes to a file"
                "called <test_story>_prefixes.txt, but there's no test_story.")

        fn = os.path.join(TGT_DIR, "%s_prefixes.txt" % self.test_story)
        with open(fn, "w") as f:
            for utter_id in self.test_prefixes:
                print(utter_id.split("/")[1], file=f)

    def __repr__(self):
        return ("%s(" % self.__class__.__name__ +
                "feat_type=\"%s\",\n" % self.feat_type +
                "\tlabel_type=\"%s\",\n" % self.label_type +
                "\ttrain_rec_type=\"%s\",\n" % self.train_rec_type +
                "\tmax_samples=%s,\n" % self.max_samples +
                "\tvalid_story=%s,\n" % repr(self.valid_story) +
                "\ttest_story=%s)\n" % repr(self.test_story))



from typing import Any, Callable, List, Sequence, TypeVar, Tuple

import numpy as np

from persephone.exceptions import EmptyReferenceException

T = TypeVar("T")

def min_edit_distance(
        source: Sequence[T], target: Sequence[T],
        ins_cost: Callable[..., int] = lambda _x: 1,
        del_cost: Callable[..., int] = lambda _x: 1,
        sub_cost: Callable[..., int] = lambda x, y: 0 if x == y else 1) -> int:
    """Calculates the minimum edit distance between two sequences.

    Uses the Levenshtein weighting as a default, but offers keyword arguments
    to supply functions to measure the costs for editing with different
    elements.

    Args:
        ins_cost: A function describing the cost of inserting a given char
        del_cost: A function describing the cost of deleting a given char
        sub_cost: A function describing the cost of substituting one char for

    Returns:
        The edit distance between the two input sequences.

    """

    # Initialize an m+1 by n+1 array. Note that the strings start from index 1,
    # with index 0 being used to denote the empty string.
    n = len(target)
    m = len(source)
    distance = np.zeros((m+1, n+1), dtype=np.int16)

    # Initialize the zeroth row and column to be the distance from the empty
    # string.
    for i in range(1, m+1):
        distance[i, 0] = distance[i-1, 0] + ins_cost(source[i-1])
    for j in range(1, n+1):
        distance[0, j] = distance[0, j-1] + ins_cost(target[j-1])

    # Do the dynamic programming to fill in the matrix with the edit distances.
    for j in range(1, n+1):
        for i in range(1, m+1):
            distance[i, j] = min(
                    distance[i-1, j] + ins_cost(source[i-1]),
                    distance[i-1, j-1] + sub_cost(source[i-1],target[j-1]),
                    distance[i, j-1] + del_cost(target[j-1]))

    return int(distance[len(source), len(target)])

def min_edit_distance_align(
        # TODO Wrangle the typing errors in this function.
        # TODO This could work on generic sequences but for now it relies on
        # empty strings.
        #source: Sequence[str], target: Sequence[str],
        #ins_cost: Callable[..., int] = lambda _x: 1,
        #del_cost: Callable[..., int] = lambda _x: 1,
        #sub_cost: Callable[..., int] = lambda x, y: 0 if x == y else 1
        #) -> List[Tuple[str, str]]:
        source, target,
        ins_cost = lambda _x: 1,
        del_cost = lambda _x: 1,
        sub_cost = lambda x, y: 0 if x == y else 1):
    """Finds a minimum cost alignment between two strings.

    Uses the Levenshtein weighting as a default, but offers keyword arguments
    to supply functions to measure the costs for editing with different
    characters. Note that the alignment may not be unique.

    Args:
        ins_cost: A function describing the cost of inserting a given char
        del_cost: A function describing the cost of deleting a given char
        sub_cost: A function describing the cost of substituting one char for

    Returns:
        A sequence of tuples representing character level alignments between
        the source and target strings.
    """

    # Initialize an m+1 by n+1 array to hold the distances, and an equal sized
    # array to store the backpointers. Note that the strings start from index
    # 1, with index 0 being used to denote the empty string.
    n = len(target)
    m = len(source)
    dist = [[0]*(n+1) for _ in range(m+1)]
    bptrs = [[[]]*(n+1) for _ in range(m+1)] # type: List[List[List]]

    # Adjust the initialization of the first column and row of the matrices to
    # their appropriate values.
    for i in range(1, m+1):
        dist[i][0] = i
        bptrs[i][0] = (i-1, 0)
    for j in range(1, n+1):
        dist[0][j] = j
        bptrs[0][j] = (0, j-1)

    # Do the dynamic programming to fill in the matrix with the edit distances.
    for j in range(1, n+1):
        for i in range(1, m+1):
            options = [
                    (dist[i-1][j] + ins_cost(target[j-1]),
                        (i-1, j)),
                    (dist[i-1][j-1] + sub_cost(source[i-1],target[j-1]),
                        (i-1, j-1)),
                    (dist[i][j-1] + del_cost(source[i-1]),
                        (i, j-1))]
            (minimum, pointer) = sorted(options)[0]
            dist[i][j] = minimum
            bptrs[i][j] = pointer

    # Put the backtrace in a list, and reverse it to get a forward trace.
    bt = [(m,n)]
    cell = bptrs[m][n]
    while True:
        if not cell:
            # If it's an empty list or tuple, as will be the case at the start
            # of the trace
            break
        bt.append(cell)
        if bptrs[cell[0]][cell[1]]:
            cell = bptrs[cell[0]][cell[1]]
        else:
            break
    trace = list(reversed(bt))

    # Construct an alignment between source and target using the trace.
    alignment = []
    for i in range(1, len(trace)):
        current = trace[i]
        prev = trace[i-1]

        # If the cell is diagonal from the previous one, it's a substitution or
        # there wasn't a change.
        if current[0] - prev[0] == 1 and current[1] - prev[1] == 1:
            alignment.append((source[current[0]-1], target[current[1]-1]))
        # Otherwise if it moves only on the source side, it's a deletion
        elif current[0] - prev[0] == 1:
            alignment.append((source[current[0]-1], ""))
        # Otherwise if it moves only on the target side, it's an insertion
        elif current[1] - prev[1] == 1:
            alignment.append(("", target[current[1]-1]))

    return alignment

# TODO Wrangle the typing errors in this function.
#def cluster_alignment_errors(alignment: List[Tuple[str, str]]
#                            ) -> List[Tuple[Tuple[str, ...]]]:
def cluster_alignment_errors(alignment):
    # TODO Review documentation and consider for inclusion in API.
    """Clusters alignments
    Takes an alignment created by min_edit_distance_align() and groups
    consecutive errors together. This is useful, because there are often
    many possible alignments, and so often we can't meaningfully distinguish
    between alignment errors at the character level, so it makes many-to-many
    mistakes more readable."""

    newalign = []
    mistakes = ([],[])
    for align_item in alignment:
        if align_item[0] == align_item[1]:
            if mistakes != ([],[]):
                newalign.append((tuple(mistakes[0]), tuple(mistakes[1])))
                mistakes = ([],[])
            newalign.append((tuple([align_item[0]]), tuple([align_item[1]])))
        else:
            if align_item[0] != "":
                mistakes[0].append(align_item[0])
            if align_item[1] != "":
                mistakes[1].append(align_item[1])
    if mistakes != ([],[]):
        newalign.append((tuple(mistakes[0]), tuple(mistakes[1])))
        mistakes = ([],[])

    return newalign

def word_error_rate(ref: Sequence[T], hyp: Sequence[T]) -> float:
    """ Calculate the word error rate of a sequence against a reference.

    Args:
        ref: The gold-standard reference sequence
        hyp: The hypothesis to be evaluated against the reference.

    Returns:
        The word error rate of the supplied hypothesis with respect to the
        reference string.

    Raises:
        persephone.exceptions.EmptyReferenceException: If the length of the reference sequence is 0.

    """

    if len(ref) == 0:
        raise EmptyReferenceException(
            "Cannot calculating word error rate against a length 0 "\
            "reference sequence.")

    distance = min_edit_distance(ref, hyp)
    return 100 * float(distance) / len(ref)



""" Performs feature extraction of WAV files for acoustic modelling."""

import logging
import os
from pathlib import Path
import subprocess
from typing import Union
import wave

import numpy as np
import python_speech_features
import scipy.io.wavfile as wav

from .. import config
from ..exceptions import PersephoneException

logger = logging.getLogger(__name__) #type: ignore

def empty_wav(wav_path: Union[Path, str]) -> bool:
    """Check if a wav contains data"""
    with wave.open(str(wav_path), 'rb') as wav_f:
        return wav_f.getnframes() == 0


def extract_energy(rate, sig):
    """ Extracts the energy of frames. """

    mfcc = python_speech_features.mfcc(sig, rate, appendEnergy=True)
    energy_row_vec = mfcc[:, 0]
    energy_col_vec = energy_row_vec[:, np.newaxis]
    return energy_col_vec

def fbank(wav_path, flat=True):
    """ Currently grabs log Mel filterbank, deltas and double deltas."""

    (rate, sig) = wav.read(wav_path)
    if len(sig) == 0:
        logger.warning("Empty wav: {}".format(wav_path))
    fbank_feat = python_speech_features.logfbank(sig, rate, nfilt=40)
    energy = extract_energy(rate, sig)
    feat = np.hstack([energy, fbank_feat])
    delta_feat = python_speech_features.delta(feat, 2)
    delta_delta_feat = python_speech_features.delta(delta_feat, 2)
    all_feats = [feat, delta_feat, delta_delta_feat]
    if not flat:
        all_feats = np.array(all_feats)
        # Make time the first dimension for easy length normalization padding
        # later.
        all_feats = np.swapaxes(all_feats, 0, 1)
        all_feats = np.swapaxes(all_feats, 1, 2)
    else:
        all_feats = np.concatenate(all_feats, axis=1)

    # Log Mel Filterbank, with delta, and double delta
    feat_fn = wav_path[:-3] + "fbank.npy"
    np.save(feat_fn, all_feats)

def mfcc(wav_path):
    """ Grabs MFCC features with energy and derivates. """

    (rate, sig) = wav.read(wav_path)
    feat = python_speech_features.mfcc(sig, rate, appendEnergy=True)
    delta_feat = python_speech_features.delta(feat, 2)
    all_feats = [feat, delta_feat]
    all_feats = np.array(all_feats)
    # Make time the first dimension for easy length normalization padding later.
    all_feats = np.swapaxes(all_feats, 0, 1)
    all_feats = np.swapaxes(all_feats, 1, 2)

    feat_fn = wav_path[:-3] + "mfcc13_d.npy"
    np.save(feat_fn, all_feats)

def combine_fbank_and_pitch(feat_dir: str, prefix: str) -> None:

    fbank_fn = os.path.join(feat_dir, prefix + ".fbank.npy")
    fbanks = np.load(fbank_fn)
    pitch_fn = os.path.join(feat_dir, prefix + ".pitch.npy")
    pitches = np.load(pitch_fn)

    # Check that the fbanks are flat
    if len(fbanks.shape) == 3:
        def flatten(feats_3d):
            swapped = np.swapaxes(feats_3d, 0, 1)
            concatenated = np.concatenate(swapped, axis=1)
            return concatenated
        fbanks = flatten(fbanks)
    elif len(fbanks.shape) == 2:
        pass
    else:
        logger.error("Invalid fbank array shape %s", (str(fbanks.shape)))
        raise PersephoneException("Invalid fbank array shape %s" % (str(fbanks.shape)))

    diff = len(fbanks) - len(pitches)

    # When using python_speech_features fbank extraction and kaldi pitch
    # extraction, there are slightly differing numbers of frames. Usually only
    # by 1 or so. Just pad with zeros to match.
    # TODO This could be a sign that it might just be best to use Kaldi for
    # fbank feature extraction as well (I kind of want to see how using those
    # features goes anyway). But I'm currently keeping it this way for
    # experimental consistency.
    if diff > 2:
        logger.warning("Excessive difference in number of frames. %d", diff)
        raise PersephoneException("Excessive difference in number of frames. %d" % diff)
    elif diff > 0:
        pitches = np.concatenate((np.array([[0,0]]*(len(fbanks) - len(pitches))), pitches))
    fbank_pitch_feats = np.concatenate((fbanks, pitches), axis=1)

    out_fn = os.path.join(feat_dir, prefix + ".fbank_and_pitch.npy")
    np.save(out_fn, fbank_pitch_feats)

# TODO I'm operating at the directory level for pitch feats, but at the file
# level for other things. Change this? Currently wav normalization occurs
# elsewhere too (in say, datasets.chatin.prepare_feats()). Kaldi expects
# specific wav format, so that should be coupled together with pitch extraction
# here.
def from_dir(dirpath: Path, feat_type: str) -> None:
    """ Performs feature extraction from the WAV files in a directory.

    Args:
        dirpath: A `Path` to the directory where the WAV files reside.
        feat_type: The type of features that are being used.
    """

    logger.info("Extracting features from directory {}".format(dirpath))

    dirname = str(dirpath)

    def all_wavs_processed() -> bool:
        """
        True if all wavs in the directory have corresponding numpy feature
        file; False otherwise.
        """

        for fn in os.listdir(dirname):
            prefix, ext = os.path.splitext(fn)
            if ext == ".wav":
                if not os.path.exists(
                        os.path.join(dirname, "%s.%s.npy" % (prefix, feat_type))):
                    return False
        return True

    if all_wavs_processed():
        # Then nothing needs to be done here
        logger.info("All WAV files already preprocessed")
        return
    # Otherwise, go on and process everything...

    # If pitch features are needed as part of this, extract them
    if feat_type == "pitch" or feat_type == "fbank_and_pitch":
        kaldi_pitch(dirname, dirname)

    # Then apply file-wise feature extraction
    for filename in os.listdir(dirname):
        logger.info("Preparing %s features for %s", feat_type, filename)
        path = os.path.join(dirname, filename)
        if path.endswith(".wav"):
            if empty_wav(path):
                raise PersephoneException("Can't extract features for {} since it is an empty WAV file. Remove it from the corpus.".format(path))
            if feat_type == "fbank":
                fbank(path)
            elif feat_type == "fbank_and_pitch":
                fbank(path)
                prefix = os.path.splitext(filename)[0]
                combine_fbank_and_pitch(dirname, prefix)
            elif feat_type == "pitch":
                # Already extracted pitch at the start of this function.
                pass
            elif feat_type == "mfcc13_d":
                mfcc(path)
            else:
                logger.warning("Feature type not found: %s", feat_type)
                raise PersephoneException("Feature type not found: %s" % feat_type)

def convert_wav(org_wav_fn: Path, tgt_wav_fn: Path) -> None:
    """ Converts the wav into a 16bit mono 16000Hz wav.

        Args:
            org_wav_fn: A `Path` to the original wave file
            tgt_wav_fn: The `Path` to output the processed wave file
    """
    if not org_wav_fn.exists():
        raise FileNotFoundError
    args = [config.FFMPEG_PATH,
            "-i", str(org_wav_fn), "-ac", "1", "-ar", "16000", str(tgt_wav_fn)]
    subprocess.run(args)

def kaldi_pitch(wav_dir: str, feat_dir: str) -> None:
    """ Extract Kaldi pitch features. Assumes 16k mono wav files."""

    logger.debug("Make wav.scp and pitch.scp files")
    # Make wav.scp and pitch.scp files
    prefixes = []
    for fn in os.listdir(wav_dir):
        prefix, ext = os.path.splitext(fn)
        if ext == ".wav":
            prefixes.append(prefix)

    wav_scp_path = os.path.join(feat_dir, "wavs.scp")
    with open(wav_scp_path, "w") as wav_scp:
        for prefix in prefixes:
            logger.info("Writing wav file: %s", os.path.join(wav_dir, prefix + ".wav"))
            print(prefix, os.path.join(wav_dir, prefix + ".wav"), file=wav_scp)

    pitch_scp_path = os.path.join(feat_dir, "pitch_feats.scp")
    with open(pitch_scp_path, "w") as pitch_scp:
        for prefix in prefixes:
            logger.info("Writing scp file: %s", os.path.join(feat_dir, prefix + ".pitch.txt"))
            print(prefix, os.path.join(feat_dir, prefix + ".pitch.txt"), file=pitch_scp)

    # Call Kaldi pitch feat extraction
    args = [os.path.join(config.KALDI_ROOT, "src/featbin/compute-kaldi-pitch-feats"),
            "scp:%s" % (wav_scp_path), "scp,t:%s" % pitch_scp_path]
    logger.info("Extracting pitch features from wavs listed in {}".format(
        wav_scp_path))
    subprocess.run(args)

    # Convert the Kaldi pitch *.txt files to numpy arrays.
    for fn in os.listdir(feat_dir):
        if fn.endswith(".pitch.txt"):
            pitch_feats = []
            with open(os.path.join(feat_dir, fn)) as f:
                for line in f:
                    sp = line.split()
                    if len(sp) > 1:
                        pitch_feats.append([float(sp[0]), float(sp[1])])
            prefix, _ = os.path.splitext(fn)
            out_fn = prefix + ".npy"
            a = np.array(pitch_feats)
            np.save(os.path.join(feat_dir, out_fn), a)



from collections import defaultdict
from pathlib import Path
from typing import List, NamedTuple, Set, Sequence, Tuple, DefaultDict, Dict

Utterance = NamedTuple("Utterance", [("org_media_path", Path),
                                     ("org_transcription_path", Path),
                                     ("prefix", str),
                                     ("start_time", int),
                                     ("end_time", int),
                                     ("text", str),
                                     ("speaker", str)])
Utterance.__doc__= (
    """ An immutable object that represents a single utterance.

    `Utterance` instances capture key data about short segments of speech in
    the corpus.  Their most important role is in representing transcriptions
    in various states of preprocessing. For instance, `Utterance` instances may
    be created when reading from a linguists transcription files, in which case
    their `text` attribute is a raw unpreprocessed transcription. These
    `Utterance` instances may then be fed to a function that preprocesses the
    text, returning new `Utterance` instances with, say, phonemes delimited
    with spaces so that they are in an appropriate format for model training.

    Note that `Utterance` instances are not required as arguments to `Corpus`
    constructors. They exist to aid in preprocessing.

    Attributes:
        org_media_path: A `pathlib.Path` to the original source audio that contains the
            utterance (which may comprise many utterances).
        org_transcription_path: A `pathlib.Path` to the source of the transcription of
            the utterance (which may comprise many utterances in the
            case of, say, ELAN files).
        prefix: A string identifier for the utterance which is used to prefix the
            target wav and transcription files, which are called `<prefix>.wav`,
            `<prefix>.phonemes`, etc.
        start_time: An integer denoting the offset, in milliseconds, of the
            utterance in the original media file found in `org_media_path`.
        end_time: An integer denoting the endpoint, in milliseconds, of the
            utterance in the original media file found in `org_media_path`.
        text: A string representation of the transcription.
        speaker: A string representation of the speaker of the utterance.

    """)

def write_transcriptions(utterances: List[Utterance],
                         tgt_dir: Path, ext: str, lazy: bool) -> None:
    """ Write the utterance transcriptions to files in the tgt_dir. Is lazy and
    checks if the file already exists.

    Args:
        utterances: A list of Utterance objects to be written.
        tgt_dir: The directory in which to write the text of the utterances,
            one file per utterance.
        ext: The file extension for the utterances. Typically something like
            "phonemes", or "phonemes_and_tones".

    """

    tgt_dir.mkdir(parents=True, exist_ok=True)
    for utter in utterances:
        out_path = tgt_dir / "{}.{}".format(utter.prefix, ext)
        if lazy and out_path.is_file():
            continue
        with out_path.open("w") as f:
            print(utter.text, file=f)

def remove_duplicates(utterances: List[Utterance]) -> List[Utterance]:
    """ Removes utterances with the same start_time, end_time and text. Other
    metadata isn't considered.
    """

    filtered_utters = []
    utter_set = set() # type: Set[Tuple[int, int, str]]
    for utter in utterances:
        if (utter.start_time, utter.end_time, utter.text) in utter_set:
            continue
        filtered_utters.append(utter)
        utter_set.add((utter.start_time, utter.end_time, utter.text))

    return filtered_utters

def remove_empty_text(utterances: List[Utterance]) -> List[Utterance]:
    """Remove empty utterances from a list of utterances
    Args:
        utterances: The list of utterance we are processing
    """
    return [utter for utter in utterances if utter.text.strip() != ""]

# Doing everything in milliseconds now; other units are only for reporting to
# users
def duration(utter: Utterance) -> int:
    """Get the duration of an utterance in milliseconds
    Args:
        utter: The utterance we are finding the duration of
    """
    return utter.end_time - utter.start_time

def total_duration(utterances: List[Utterance]) -> int:

    """Get the duration of an entire list of utterances in milliseconds
    Args:
        utterances: The list of utterance we are finding the duration of
    """
    return sum([duration(utter) for utter in utterances])

def make_speaker_utters(utterances: List[Utterance]) -> Dict[str, List[Utterance]]:
    """ Creates a dictionary mapping from speakers to their utterances. """

    speaker_utters = defaultdict(list) # type: DefaultDict[str, List[Utterance]]
    for utter in utterances:
        speaker_utters[utter.speaker].append(utter)

    return speaker_utters

def speaker_durations(utterances: List[Utterance]) -> List[Tuple[str, int]]:
    """ Takes a list of utterances and itemizes them by speaker, returning a
    list of tuples of the form (Speaker Name, duration).
    """

    speaker_utters = make_speaker_utters(utterances)

    speaker_duration_tuples = [] # type: List[Tuple[str, int]]
    for speaker in speaker_utters:
        speaker_duration_tuples.append((speaker, total_duration(speaker_utters[speaker])))

    return speaker_duration_tuples

def remove_too_short(utterances: List[Utterance],
                     _winlen=25, winstep=10) -> List[Utterance]:
    """ Removes utterances that will probably have issues with CTC because of
    the number of frames being less than the number of tokens in the
    transcription. Assuming char tokenization to minimize false negatives.
    """
    def is_too_short(utterance: Utterance) -> bool:
        charlen = len(utterance.text)
        if (duration(utterance) / winstep) < charlen:
            return True
        else:
            return False

    return [utter for utter in utterances if not is_too_short(utter)]



"""Context manager for changing the current working directory"""

import os


class cd:
    """Context manager for changing the current working directory"""
    def __init__(self, new_path):
        self.new_path = os.path.expanduser(str(new_path))

    def __enter__(self):
        self.saved_path = os.getcwd()
        os.chdir(self.new_path)

    def __exit__(self, etype, value, traceback):
        os.chdir(self.saved_path)



""" Interface to Alex/Steven's Kunwinjku data. """

from pathlib import Path
import subprocess
from typing import List, Set

import nltk # type: ignore
# TODO This download should be conditional, since a complaint is raised if
# there is no net connection
nltk.download("punkt") # type: ignore
from pympi.Elan import Eaf

from .. import corpus
from .. import config
from ..preprocess.labels import segment_into_tokens
from ..utterance import Utterance
from ..preprocess.labels import LabelSegmenter
from ..corpus import Corpus

BASIC_PHONEMES = set(["a", "b", "d", "dj", "rd", "e", "h", "i", "k", "l",
            "rl", "m", "n", "ng", "nj", "rn", "o", "r", "rr", "u",
            "w", "y",])
DOUBLE_STOPS = set(["bb", "dd", "djdj", "rdd", "kk"])
DIPHTHONGS = set(["ay", "aw", "ey", "ew", "iw", "oy", "ow", "uy"])
PHONEMES = BASIC_PHONEMES | DOUBLE_STOPS | DIPHTHONGS

def pull_en_words() -> None:
    """ Fetches a repository containing English words. """

    ENGLISH_WORDS_URL = "https://github.com/dwyl/english-words.git"
    en_words_path = Path(config.EN_WORDS_PATH)
    if not en_words_path.is_file():
        subprocess.run(["git", "clone",
                        ENGLISH_WORDS_URL, str(en_words_path.parent)])

def get_en_words() -> Set[str]:
    """
    Returns a list of English words which can be used to filter out
    code-switched sentences.
    """

    pull_en_words()

    with open(config.EN_WORDS_PATH) as words_f:
        raw_words = words_f.readlines()
    en_words = set([word.strip().lower() for word in raw_words])
    NA_WORDS_IN_EN_DICT = set(["kore", "nani", "karri", "imi", "o", "yaw", "i",
                           "bi", "aye", "imi", "ane", "kubba", "kab", "a-",
                           "ad", "a", "mak", "selim", "ngai", "en", "yo",
                           "wud", "mani", "yak", "manu", "ka-", "mong",
                           "manga", "ka-", "mane", "kala", "name", "kayo",
                           "kare", "laik", "bale", "ni", "rey", "bu",
                           "re", "iman", "bom", "wam",
                           "alu", "nan", "kure", "kuri", "wam", "ka", "ng",
                           "yi", "na", "m", "arri", "e", "kele", "arri", "nga",
                           "kakan", "ai", "ning", "mala", "ti", "wolk",
                           "bo", "andi", "ken", "ba", "aa", "kun", "bini",
                           "wo", "bim", "man", "bord", "al", "mah", "won",
                           "ku", "ay", "belen", "wen", "yah", "muni",
                           "bah", "di", "mm", "anu", "nane", "ma", "kum",
                           "birri", "ray", "h", "kane", "mumu", "bi", "ah",
                           "i-", "n", "mi", "bedman", "rud", "le", "babu",
                           "da", "kakkak", "yun", "ande", "naw", "kam", "bolk",
                           "woy", "u", "bi-",
                           ])
    EN_WORDS_NOT_IN_EN_DICT = set(["screenprinting"])
    en_words = en_words.difference(NA_WORDS_IN_EN_DICT)
    en_words = en_words | EN_WORDS_NOT_IN_EN_DICT
    return en_words

EN_WORDS = get_en_words()

def explore_elan_files(elan_paths):
    """
    A function to explore the tiers of ELAN files.
    """

    for elan_path in elan_paths:
        print(elan_path)
        eafob = Eaf(elan_path)
        tier_names = eafob.get_tier_names()
        for tier in tier_names:
            print("\t", tier)
            try:
                for annotation in eafob.get_annotation_data_for_tier(tier):
                    print("\t\t", annotation)
            except KeyError:
                continue

        input()

def segment_utterance(utterance: Utterance) -> Utterance:
    fields = utterance._asdict()
    fields["text"] = segment_str(fields["text"])
    return Utterance(**fields)

def segment_str(text: str, phoneme_inventory: Set[str] = PHONEMES) -> str:
    """
    Takes as input a string in Kunwinjku and segments it into phoneme-like
    units based on the standard orthographic rules specified at
    http://bininjgunwok.org.au/
    """

    text = text.lower()
    text = segment_into_tokens(text, phoneme_inventory)
    return text

bkw_label_segmenter = LabelSegmenter(segment_utterance, PHONEMES)

def explore_code_switching(utterances: List[Utterance], out_path: Path) -> None:

    with out_path.open("w") as f:
        en_count = 0
        for i, utter in enumerate(utterances):
            toks = nltk.word_tokenize(utter.text) # type: ignore
            words = [tok.lower() for tok in toks]
            for word in words:
                if word in EN_WORDS:
                    en_count += 1
                    print("Prefix: {}".format(utter.prefix), file=f)
                    print("Utterance #%s" % i, file=f)
                    print("Original: %s" % utter.text, file=f)
                    print("Tokenized: %s" % words, file=f)
                    print("Phonemic: %s" % segment_str(utter.text), file=f)
                    print("En word: %s" % word, file=f)
                    print("---------------------------------------------", file=f)
                    break
        print("Num of codeswitched/English sentences: {}".format(en_count), file=f)
        print("Total number of utterances: {}".format(len(utterances)), file=f)

def filter_for_not_codeswitched(utter: Utterance) -> bool:
    toks = nltk.word_tokenize(utter.text) # type: ignore
    words = [tok.lower() for tok in toks]
    for word in words:
        if word in EN_WORDS:
            return False
    return True

def filter_for_not_empty(utter: Utterance) -> bool:
    return not utter.text.strip() == ""

def bkw_filter(utter: Utterance) -> bool:
    return filter_for_not_codeswitched(utter) and filter_for_not_empty(utter)

def create_corpus(org_dir: Path = Path(config.BKW_PATH),
                 tgt_dir: Path = Path(config.TGT_DIR, "BKW"),
                 feat_type: str = "fbank", label_type: str = "phonemes",
                 speakers: List[str] = None) -> Corpus:

    if label_type != "phonemes":
        raise NotImplementedError(
            "label_type {} not implemented.".format(label_type))

    # If there is a way to sub-divide utterances, the BKW data would increase
    # significantly. This from_elan call will result in quite a few filtered
    # utterances.

    return corpus.Corpus.from_elan(org_dir, tgt_dir,
                     feat_type=feat_type, label_type=label_type,
                     utterance_filter=bkw_filter,
                     label_segmenter=bkw_label_segmenter,
                     speakers=speakers, tier_prefixes=("xv", "rf"))



"""
Describes the abstract Corpus class that all interfaces to corpora should
subclass.
"""

from collections import namedtuple
import logging.config
import os
from pathlib import Path
import pickle
import random
import subprocess
from typing import Any, List, Callable, Optional, Set, Sequence, Tuple, Type, TypeVar

import numpy as np

from .config import ENCODING
from .preprocess import feat_extract
from . import utils
from .exceptions import PersephoneException
from .exceptions import LabelMismatchException
from .preprocess import elan, wav
from . import utterance
from .utterance import Utterance
from .preprocess.labels import LabelSegmenter

logger = logging.getLogger(__name__) # type: ignore

CorpusT = TypeVar("CorpusT", bound="Corpus")

def ensure_no_set_overlap(train: Sequence[str], valid: Sequence[str], test: Sequence[str]) -> None:
    """ Ensures no test set data has creeped into the training set."""

    logger.debug("Ensuring that the training, validation and test data sets have no overlap")
    train_s = set(train)
    valid_s = set(valid)
    test_s = set(test)

    if train_s & valid_s:
        logger.warning("train and valid have overlapping items: {}".format(train_s & valid_s))
        raise PersephoneException("train and valid have overlapping items: {}".format(train_s & valid_s))
    if train_s & test_s:
        logger.warning("train and test have overlapping items: {}".format(train_s & test_s))
        raise PersephoneException("train and test have overlapping items: {}".format(train_s & test_s))
    if valid_s & test_s:
        logger.warning("valid and test have overlapping items: {}".format(valid_s & test_s))
        raise PersephoneException("valid and test have overlapping items: {}".format(valid_s & test_s))

def find_untranscribed_wavs(wav_path: Path, transcription_path: Path, label_type: str) -> List[str]:
    """Find the prefixes for all the wav files that do not have an associated transcription
    Args:
        wav_path: Path to search for wav files in
        transcription_path: Path to search for transcriptions in
        label_type: The type of labels for transcriptions. Eg "phonemes" "phonemes_and_tones"
    Returns:
        A list of all untranscribed prefixes
    """
    audio_files = wav_path.glob("**/*.wav")
    transcription_files = transcription_path.glob("**/*.{}".format(label_type))

    transcription_file_prefixes = [t_file.stem for t_file in transcription_files]

    untranscribed_prefixes = [] # type: List[str]
    for a_file in audio_files:
        if a_file.stem not in transcription_file_prefixes:
            untranscribed_prefixes.append(a_file.stem)
    return untranscribed_prefixes

def get_untranscribed_prefixes_from_file(target_directory: Path) -> List[str]:
    """
    The file "untranscribed_prefixes.txt" will specify prefixes which
    do not have an associated transcription file if placed in the target directory.

    This will fetch those prefixes from that file and will return an empty
    list if that file does not exist.

    See find_untranscribed_wavs function for finding untranscribed prefixes in an
    experiment directory.

    Returns:
        A list of all untranscribed prefixes as specified in the file
    """

    untranscribed_prefix_fn = target_directory / "untranscribed_prefixes.txt"
    if untranscribed_prefix_fn.exists():
        with untranscribed_prefix_fn.open() as f:
            prefixes = f.readlines()

        return [prefix.strip() for prefix in prefixes]
    else:
        #logger.warning("Attempting to get untranscribed prefixes but the file ({})"
        #                " that should specify these does not exist".format(untranscribed_prefix_fn))
        pass
    return []

class Corpus:
    """ Represents a preprocessed corpus that is ready to be used in model
    training.

    Construction of a `Corpus` instance involves preprocessing data if the data has
    not previously already been preprocessed. The extent of the preprocessing
    depends on which constructor is used. If the default constructor,
    `__init__()` is used, transcriptions are assumed to already be preprocessed
    and only speech feature extraction from WAV files is performed. In other
    constructors such as `from_elan()`, preprocessing of the transcriptions is
    performed. See the documentation of the relevant constructors for more
    information.

    Once a Corpus object is created it should be considered immutable. At this
    point feature extraction from WAVs will have been performed, with feature
    files in `tgt_dir/feat/`.  Transcriptions will have been segmented into
    appropriate tokens (labels) and will be stored in `tgt_dir/label/`.

    """

    def __init__(self, feat_type: str, label_type: str, tgt_dir: Path,
                 *,
                 labels: Optional[Set[str]] = None,
                 max_samples: int=1000,
                 speakers: Optional[Sequence[str]] = None) -> None:
        """ Construct a `Corpus` instance from preprocessed data.

        Assumes that the corpus data has been preprocessed and is
        structured as follows: (1) WAVs for each utterance are found in
        `<tgt_dir>/wav/` with the filename `<prefix>.wav`, where `prefix` is
        some string uniquely identifying the utterance; (2) For each WAV file,
        there is a corresponding transcription found in `<tgt_dir>/label/` with
        the filename `<prefix>.<label_type>`, where `label_type` is some string
        describing the type of label used (for example, "phonemes" or "tones").

        If the data is found in the format, WAV normalization and speech
        feature extraction will be performed during `Corpus` construction, and
        the utterances will be randomly divided into training, validation and
        test_sets. If you would like to define these datasets yourself, include
        files named `train_prefixes.txt`, `valid_prefixes.txt` and
        `test_prefixes.txt` in `<tgt_dir>`. Each file should be a list of
        prefixes (utterance IDs), one per line. If these are found during
        `Corpus` construction, those sets will be used instead.

        Args:
            feat_type: A string describing the input speech features. For
                       example, "fbank" for log Mel filterbank features.
            label_type: A string describing the transcription labels. For example,
                         "phonemes" or "tones".
            labels: A set of strings representing labels (tokens) used in
                    transcription. For example: {"a", "o", "th", ...}.
                    If this parameter is not provided the experiment directory is
                    scanned for labels present in the transcription files.
            max_samples: The maximum number of samples an utterance in the
                corpus may have. If an utterance is longer than this, it is not
                included in the corpus.

        """

        if speakers:
            raise NotImplementedError("Speakers not implemented")

        logger.debug("Creating a new Corpus object with feature type {}, label type {},"
                     "target directory {}, label set {}, max_samples {}, speakers {}".format(
                    feat_type, label_type, labels, tgt_dir, max_samples, speakers)
        )

        # In case path is supplied as a string, make it a Path
        self.tgt_dir = Path(tgt_dir)

        #: A string representing the type of speech feature (eg. "fbank"
        #: for log filterbank energies).
        self.feat_type = feat_type

        #: An arbitrary string representing the transcription tokenization
        #: used (eg. "phonemes", "tones", "joint", or "characters").
        self.label_type = label_type

        # Setting up directories
        # Set the directory names
        self.feat_dir = self.get_feat_dir()
        self.wav_dir = self.get_wav_dir()
        self.label_dir = self.get_label_dir()

        logger.debug("Setting up directories for this Corpus object at %s", self.tgt_dir)
        self.set_and_check_directories(self.tgt_dir)

        # Label-related stuff
        if labels is not None:
            self.labels = labels
            found_labels = determine_labels(self.tgt_dir, label_type)
            if found_labels != self.labels:
                raise LabelMismatchException("User specified labels, {}, do"
                    " not match those automatically found, {}.".format(labels,
                    found_labels))
        else:
            self.labels = determine_labels(self.tgt_dir, label_type)
        self.vocab_size = len(self.labels)
        self.LABEL_TO_INDEX, self.INDEX_TO_LABEL = self.initialize_labels(self.labels)
        logger.info("Corpus label set: \n\t{}".format(self.labels))

        # This is a lazy function that assumes wavs are already in the WAV dir
        # but only creates features if necessary
        logger.debug("Preparing features")
        self.prepare_feats()
        self._num_feats = None

        self.train_prefixes = [] # type: List[str]
        self.valid_prefixes = [] # type: List[str]
        self.test_prefixes = [] # type: List[str]
        # This is also lazy if the {train,valid,test}_prefixes.txt files exist.
        self.make_data_splits(max_samples=max_samples)

        # Sort the training prefixes by size for more efficient training
        logger.debug("Training prefixes")
        self.train_prefixes = utils.sort_by_size(
            self.feat_dir, self.train_prefixes, feat_type)

        # Ensure no overlap between training and test sets
        try:
            ensure_no_set_overlap(
                self.get_train_fns()[0],
                self.get_valid_fns()[0],
                self.get_test_fns()[0]
            )
        except PersephoneException:
            logger.error("Got overlap between train, valid and test data sets")
            raise

        untranscribed_from_file = self.get_untranscribed_prefixes()
        untranscribed_found = find_untranscribed_wavs(self.get_wav_dir(), self.get_label_dir(), self.label_type)

        self.untranscribed_prefixes = list(set(untranscribed_from_file) | set(untranscribed_found))

        # TODO Need to contemplate whether Corpus objects have Utterance
        # objects or # not. Some of the TestBKW tests currently rely on this
        # for testing.
        self.utterances = []  # type: List[Utterance]

        self.pickle()

    @classmethod
    def from_elan(cls: Type[CorpusT], org_dir: Path, tgt_dir: Path,
                  feat_type: str = "fbank", label_type: str = "phonemes",
                  *,
                  utterance_filter: Callable[[Utterance], bool] = None,
                  label_segmenter: Optional[LabelSegmenter] = None,
                  speakers: List[str] = None, lazy: bool = True,
                  tier_prefixes: Tuple[str, ...] = ("xv", "rf")) -> CorpusT:
        """ Construct a `Corpus` from ELAN files.

        Args:
            org_dir: A path to the directory containing the unpreprocessed
                data.
            tgt_dir: A path to the directory where the preprocessed data will
                be stored.
            feat_type: A string describing the input speech features. For
                       example, "fbank" for log Mel filterbank features.
            label_type: A string describing the transcription labels. For example,
                         "phonemes" or "tones".
            utterance_filter: A function that returns False if an utterance
                should not be included in the corpus and True otherwise. This
                can be used to remove undesirable utterances for training, such as
                codeswitched utterances.
            label_segmenter: An object that has an attribute `segment_labels`,
                which is creates new `Utterance` instances from old ones,
                by segmenting the tokens in their `text` attribute. Note,
                `LabelSegmenter` might be better as a function, the only issue
                is it needs to carry with it a list of labels. This could
                potentially be a function attribute.
            speakers: A list of speakers to filter for. If `None`, utterances
                from all speakers are included.
            tier_prefixes: A collection of strings that prefix ELAN tiers to
                filter for. For example, if this is `("xv", "rf")`, then tiers
                named "xv", "xv@Mark", "rf@Rose" would be extracted if they
                existed.

        """
        # This currently bails out if label_segmenter is not provided
        if not label_segmenter:
            raise ValueError("A label segmenter must be provided via label_segmenter")

        # In case path is supplied as a string, make it a Path
        if isinstance(tgt_dir, str):
            tgt_dir = Path(tgt_dir)

        # Read utterances from org_dir.
        utterances = elan.utterances_from_dir(org_dir,
                                              tier_prefixes=tier_prefixes)

        # Filter utterances based on some criteria (such as codeswitching).
        if utterance_filter:
            utterances = [utter for utter in utterances if utterance_filter(utter)]
        utterances = utterance.remove_duplicates(utterances)

        # Segment the labels in the utterances appropriately
        if label_segmenter:
            utterances = [label_segmenter.segment_labels(utter) for utter in utterances]

        # Remove utterances without transcriptions.
        utterances = utterance.remove_empty_text(utterances)

        # Remove utterances with exceptionally short wav_files that are too
        # short for CTC to work.
        utterances = utterance.remove_too_short(utterances)

        tgt_dir.mkdir(parents=True, exist_ok=True)

        # TODO A lot of these methods aren't ELAN-specific. preprocess.elan was
        # only used to get the utterances. There could be another Corpus
        # factory method that takes Utterance objects. the fromElan and
        # fromPangloss constructors could call this.

        # Writes the transcriptions to the tgt_dir/label/ dir
        utterance.write_transcriptions(utterances, (tgt_dir / "label"),
                               label_type, lazy=lazy)
        # Extracts utterance level WAV information from the input file.
        wav.extract_wavs(utterances, (tgt_dir / "wav"), lazy=lazy)

        corpus = cls(feat_type, label_type, tgt_dir,
                     labels=label_segmenter.labels, speakers=speakers)
        corpus.utterances = utterances
        return corpus

    def get_wav_dir(self) -> Path:
        return self.tgt_dir / "wav"

    def get_feat_dir(self) -> Path:
        return self.tgt_dir / "feat"

    def get_label_dir(self) -> Path:
        return self.tgt_dir / "label"

    @property
    def train_prefix_fn(self) -> Path:
        return self.tgt_dir / "train_prefixes.txt"

    @property
    def valid_prefix_fn(self) -> Path:
        return self.tgt_dir / "valid_prefixes.txt"

    @property
    def test_prefix_fn(self) -> Path:
        return self.tgt_dir / "test_prefixes.txt"

    def set_and_check_directories(self, tgt_dir: Path) -> None:
        """
        Make sure that the required directories exist in the target directory.
        set variables accordingly.
        """
        logger.info("Setting up directories for corpus in %s", tgt_dir)

        # Check directories exist.
        if not tgt_dir.is_dir():
            raise FileNotFoundError(
                "The directory {} does not exist.".format(tgt_dir))
        if not self.wav_dir.is_dir():
            raise PersephoneException(
                "The supplied path requires a 'wav' subdirectory.")
        self.feat_dir.mkdir(parents=True, exist_ok=True)
        if not self.label_dir.is_dir():
            raise PersephoneException(
                "The supplied path requires a 'label' subdirectory.")

    def initialize_labels(self, labels: Set[str]) -> Tuple[dict, dict]:
        """Create mappings from label to index and index to label"""
        logger.debug("Creating mappings for labels")

        label_to_index = {label: index for index, label in enumerate(
                                 ["pad"] + sorted(list(labels)))}
        index_to_label = {index: phn for index, phn in enumerate(
                                 ["pad"] + sorted(list(labels)))}

        return label_to_index, index_to_label

    def prepare_feats(self) -> None:
        """ Prepares input features"""

        logger.debug("Preparing input features")
        self.feat_dir.mkdir(parents=True, exist_ok=True)

        should_extract_feats = False
        for path in self.wav_dir.iterdir():
            if not path.suffix == ".wav":
                logger.info("Non wav file found in wav directory: %s", path)
                continue
            prefix = os.path.basename(os.path.splitext(str(path))[0])
            mono16k_wav_path = self.feat_dir / "{}.wav".format(prefix)
            feat_path = self.feat_dir / "{}.{}.npy".format(prefix, self.feat_type)
            if not feat_path.is_file():
                # Then we should extract feats
                should_extract_feats = True
                if not mono16k_wav_path.is_file():
                    feat_extract.convert_wav(path, mono16k_wav_path)

        # TODO Should be extracting feats on a per-file basis. Right now we
        # check if any feats files don't exist and then do all the feature
        # extraction.
        if should_extract_feats:
            feat_extract.from_dir(self.feat_dir, self.feat_type)

    def make_data_splits(self, max_samples: int) -> None:
        """ Splits the utterances into training, validation and test sets."""

        train_f_exists = self.train_prefix_fn.is_file()
        valid_f_exists = self.valid_prefix_fn.is_file()
        test_f_exists = self.test_prefix_fn.is_file()

        if train_f_exists and valid_f_exists and test_f_exists:
            logger.debug("Split for training, validation and tests specified by files")
            self.train_prefixes = self.read_prefixes(self.train_prefix_fn)
            self.valid_prefixes = self.read_prefixes(self.valid_prefix_fn)
            self.test_prefixes = self.read_prefixes(self.test_prefix_fn)
            return

        # Otherwise we now need to load prefixes for other cases addressed
        # below
        prefixes = self.determine_prefixes()
        prefixes = utils.filter_by_size(
            self.feat_dir, prefixes, self.feat_type, max_samples)

        if not train_f_exists and not valid_f_exists and not test_f_exists:
            logger.debug("No files supplied to define the split for training, validation"
                         " and tests. Using default.")
            train_prefixes, valid_prefixes, test_prefixes = self.divide_prefixes(prefixes)
            self.train_prefixes = train_prefixes
            self.valid_prefixes = valid_prefixes
            self.test_prefixes = test_prefixes
            self.write_prefixes(train_prefixes, self.train_prefix_fn)
            self.write_prefixes(valid_prefixes, self.valid_prefix_fn)
            self.write_prefixes(test_prefixes, self.test_prefix_fn)
        elif not train_f_exists and valid_f_exists and test_f_exists:
            # Then we just make all other prefixes training prefixes.
            self.valid_prefixes = self.read_prefixes(self.valid_prefix_fn)
            self.test_prefixes = self.read_prefixes(self.test_prefix_fn)
            train_prefixes = list(
                set(prefixes) - set(self.valid_prefixes))
            self.train_prefixes = list(
                set(train_prefixes) - set(self.test_prefixes))
            self.write_prefixes(self.train_prefixes, self.train_prefix_fn)
        else:
            raise NotImplementedError(
                "The following case has not been implemented:" + 
                "{} exists - {}\n".format(self.train_prefix_fn, train_f_exists) +
                "{} exists - {}\n".format(self.valid_prefix_fn, valid_f_exists) +
                "{} exists - {}\n".format(self.test_prefix_fn, test_f_exists))

    @staticmethod
    def read_prefixes(prefix_fn: Path) -> List[str]:
        if not prefix_fn.is_file():
            logger.critical("Expected a prefix file at path {}, but this path is"
                           " not a file".format(prefix_fn))
        assert prefix_fn.is_file(), "Path {} was not a file".format(prefix_fn)
        with prefix_fn.open() as prefix_f:
            prefixes = [line.strip() for line in prefix_f]
        if not prefixes:
            raise PersephoneException(
                "Empty prefix file {}. Either delete it\
                or put something in it".format(prefix_fn))
        return prefixes

    @staticmethod
    def write_prefixes(prefixes: List[str], prefix_fn: Path) -> None:
        if not prefixes:
            raise PersephoneException(
                "No prefixes. Will not write {}".format(prefix_fn))
        with prefix_fn.open("w") as prefix_f:
            for prefix in prefixes:
                print(prefix, file=prefix_f)

    @staticmethod
    def divide_prefixes(prefixes: List[str], *, seed:int=0) -> Tuple[List[str], List[str], List[str]]:
        """Divide data into training, validation and test subsets"""
        if len(prefixes) < 3:
            raise PersephoneException(
                "{} cannot be split into 3 groups as it only has {} items".format(prefixes, len(prefixes))
            )
        Ratios = namedtuple("Ratios", ["train", "valid", "test"])
        ratios=Ratios(.90, .05, .05)
        train_end = int(ratios.train*len(prefixes))
        valid_end = int(train_end + ratios.valid*len(prefixes))

        # We must make sure that at least one element exists in test
        if valid_end == len(prefixes):
            valid_end -= 1

        # If train_end and valid_end are the same we end up with no valid_prefixes
        # so we must ensure at least one prefix is placed in this category
        if train_end == valid_end:
            train_end -= 1

        random.seed(seed)
        random.shuffle(prefixes)

        train_prefixes = prefixes[:train_end]
        valid_prefixes = prefixes[train_end:valid_end]
        test_prefixes = prefixes[valid_end:]

        assert train_prefixes, "Got empty set for training data"
        assert valid_prefixes, "Got empty set for validation data"
        assert test_prefixes, "Got empty set for testing data"

        return train_prefixes, valid_prefixes, test_prefixes

    def indices_to_labels(self, indices: Sequence[int]) -> List[str]:
        """ Converts a sequence of indices into their corresponding labels."""

        return [(self.INDEX_TO_LABEL[index]) for index in indices]

    def labels_to_indices(self, labels: Sequence[str]) -> List[int]:
        """ Converts a sequence of labels into their corresponding indices."""

        return [self.LABEL_TO_INDEX[label] for label in labels]

    @property
    def num_feats(self):
        """ The number of features per time step in the corpus. """
        if not self._num_feats:
            filename = self.get_train_fns()[0][0]
            feats = np.load(filename)
            # pylint: disable=maybe-no-member
            if len(feats.shape) == 3:
                # Then there are multiple channels of multiple feats
                self._num_feats = feats.shape[1] * feats.shape[2]
            elif len(feats.shape) == 2:
                # Otherwise it is just of shape time x feats
                self._num_feats = feats.shape[1]
            else:
                raise ValueError(
                    "Feature matrix of shape %s unexpected" % str(feats.shape))
        return self._num_feats

    def prefixes_to_fns(self, prefixes: List[str]) -> Tuple[List[str], List[str]]:
        """ Fetches the file paths to the features files and labels files
        corresponding to the provided list of features"""
        # TODO Return pathlib.Paths
        feat_fns = [str(self.feat_dir / ("%s.%s.npy" % (prefix, self.feat_type)))
                    for prefix in prefixes]
        label_fns = [str(self.label_dir / ("%s.%s" % (prefix, self.label_type)))
                      for prefix in prefixes]
        return feat_fns, label_fns

    def get_train_fns(self) -> Tuple[List[str], List[str]]:
        """ Fetches the training set of the corpus.

        Outputs a Tuple of size 2, where the first element is a list of paths
        to input features files, one per utterance. The second element is a list
        of paths to the transcriptions.
        """
        return self.prefixes_to_fns(self.train_prefixes)

    def get_valid_fns(self) -> Tuple[List[str], List[str]]:
        """ Fetches the validation set of the corpus."""
        return self.prefixes_to_fns(self.valid_prefixes)

    def get_test_fns(self) -> Tuple[List[str], List[str]]:
        """ Fetches the test set of the corpus."""
        return self.prefixes_to_fns(self.test_prefixes)

    def get_untranscribed_prefixes(self) -> List[str]:
        """
        The file "untranscribed_prefixes.txt" will specify prefixes which
        do not have an associated transcription file if placed in the target directory.

        This will fetch those prefixes from that file and will return an empty
        list if that file does not exist.

        See find_untranscribed_wavs function for finding untranscribed prefixes in an
        experiment directory.
        """
        return get_untranscribed_prefixes_from_file(self.tgt_dir)

    def get_untranscribed_fns(self) -> List[str]:
        feat_fns = [os.path.join(str(self.feat_dir), "%s.%s.npy" % (prefix, self.feat_type))
                    for prefix in self.untranscribed_prefixes]
        return feat_fns

    def determine_prefixes(self) -> List[str]:
        label_prefixes = [str(path.relative_to(self.label_dir).with_suffix(""))
                          for path in 
                          self.label_dir.glob("**/*.{}".format(self.label_type))]
        wav_prefixes = [str(path.relative_to(self.wav_dir).with_suffix(""))
                          for path in 
                          self.wav_dir.glob("**/*.{}".format("wav"))]

        # Take the intersection; sort for determinism.
        prefixes = sorted(list(set(label_prefixes) & set(wav_prefixes)))

        if not prefixes:
            raise PersephoneException("""WARNING: Corpus object has no data. Are you sure
            it's in the correct directories? WAVs should be in {} and
            transcriptions in {} with the extension .{}""".format(
                self.wav_dir, self.label_dir, self.label_type))

        return prefixes

    def review(self) -> None:
        """ Used to play the WAV files and compare with the transcription. """

        for prefix in self.determine_prefixes():
            print("Utterance: {}".format(prefix))
            wav_fn = self.feat_dir / "{}.wav".format(prefix)
            label_fn = self.label_dir / "{}.{}".format(prefix,self.label_type)
            with label_fn.open() as f:
                transcript = f.read().strip()
            print("Transcription: {}".format(transcript))
            subprocess.run(["play", str(wav_fn)])

    def pickle(self) -> None:
        """ Pickles the Corpus object in a file in tgt_dir. """

        pickle_path = self.tgt_dir / "corpus.p"
        logger.debug("pickling %r object and saving it to path %s", self, pickle_path)
        with pickle_path.open("wb") as f:
            pickle.dump(self, f)

    @classmethod
    def from_pickle(cls: Type[CorpusT], tgt_dir: Path) -> CorpusT:
        pickle_path = tgt_dir / "corpus.p"
        logger.debug("Creating Corpus object from pickle file path %s", pickle_path)
        with pickle_path.open("rb") as f:
            return pickle.load(f)


def determine_labels(target_dir: Path, label_type: str) -> Set[str]:
    """ Returns a set of all phonemes found in the corpus. Assumes that WAV files and
    label files are split into utterances and segregated in a directory which contains a
    "wav" subdirectory and "label" subdirectory.

    Arguments:
        target_dir: A `Path` to the directory where the corpus data is found
        label_type: The type of label we are creating the label set from. For example
                    "phonemes" would only search for labels for that type.
    """
    logger.info("Finding phonemes of type %s in directory %s", label_type, target_dir)

    label_dir = target_dir / "label/"
    if not label_dir.is_dir():
        raise FileNotFoundError(
            "The directory {} does not exist.".format(target_dir))

    phonemes = set() # type: Set[str]
    for fn in os.listdir(str(label_dir)):
        if fn.endswith(str(label_type)):
            with (label_dir / fn).open("r", encoding=ENCODING) as f:
                try:
                    line_phonemes = set(f.readline().split())
                except UnicodeDecodeError:
                    logger.error("Unicode decode error on file %s", fn)
                    print("Unicode decode error on file {}".format(fn))
                    raise
                phonemes = phonemes.union(line_phonemes)
    return phonemes



from setuptools import setup

setup(name='persephone',
      version='0.4.1',
      description='A tool for developing automatic phoneme transcription models',
      long_description=open('README.rst', encoding="utf8").read(),
      url='https://github.com/oadams/persephone',
      author='Oliver Adams',
      author_email='oliver.adams@gmail.com',
      license='GPLv3',
      packages=['persephone', 'persephone.datasets', 'persephone.preprocess'],
      classifiers = [
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
      ],
      keywords='speech-recognition machine-learning acoustic-models artificial-intelligence neural-networks',
      install_requires=[
           'GitPython==2.1.8',
           'nltk==3.4.5',
           'numpy>=1.14.5,<2',
           'python-speech-features==0.6',
           'scipy>=1.1.0,<2',
           'tensorflow>=1.13.1,<2',
           'scikit-learn==0.21.2',
           'pympi-ling==1.69',
           'pydub==0.20.0',
           'pint==0.9',
      ],
      include_package_data = True,
)



"""Exceptions that can be raised by the Persephone library are defined here"""

class PersephoneException(Exception):
    """Base class for all exceptions raised by the Persephone library"""

class NoPrefixFileException(PersephoneException):
    """
    Thrown if files like train_prefixes.txt, test_prefixes.txt can't be
    found.
    """

class DirtyRepoException(PersephoneException):
    """
    An exception that is raised if the current working directory
    is in a dirty state according to Git.
    """

class EmptyReferenceException(PersephoneException):
    """
    When calculating word error rates, the reference string must be of
    length >= 1. Otherwise, this exception will be thrown.
    """

class LabelMismatchException(PersephoneException):
    """
    When user supplied labels are inconsistent with those automatically
    determined.
    """
